/*
 *  Definitions for macros and manifest constants used in the compiler
 *  interpreter.
 */

/*
 *  Definitions common to the compiler and interpreter.
 */

/*
 * Constants that are not likely to vary between implementations.
 */

#define BitOffMask (IntBits-1)
#define CsetSize (256/IntBits)	/* number of ints to hold 256 cset
				 *  bits. Use (256/IntBits)+1 if
				 *  256 % IntBits != 0 */
#define MinListSlots	    8	/* number of elements in an expansion
				 * list element block  */

#define MaxCvtLen	   257	/* largest string in conversions; the extra
				 *  one is for a terminating null */
#define MaxReadStr	   512	/* largest string to read() in one piece */
#define MaxIn		  32767	/* largest number of bytes to read() at once */
#define RandA        1103515245	/* random seed multiplier */
#define RandC	      453816694	/* random seed additive constant */
#define RanScale 4.65661286e-10	/* random scale factor = 1/(2^31-1)) */

/*
 * File status flags in status field of file blocks.
 */
#define Fs_Read		 01	/* read access */
#define Fs_Write	 02	/* write access */
#define Fs_Create	 04	/* file created on open */
#define Fs_Append	010	/* append mode */
#define Fs_Pipe		020	/* reading/writing on a pipe */

#ifdef RecordIO
#define Fs_Record       040     /* record structured file */
#endif					/* RecordIO */

#ifdef StandardLib
#define Fs_Reading     0100     /* last file operation was read */
#define Fs_Writing     0200     /* last file operation was write */
#endif					/* StandardLib */


/*
 * Memory sizing. 
 */

#ifdef FixedRegions
#undef IconAlloc
#ifndef AlcMax
#define AlcMax 25
#endif					/* AlcMax */
#endif					/* FixedRegions */

/*
 * Maximum sized block that can be allocated (via malloc() or such).
 */

#ifndef MaxBlock
#define MaxBlock MaxUnsigned
#endif					/* MaxBlock */

/*
 * What follows is default memory sizing. Implementations with special
 *  requirements may specify these values in define.h.
 */

#ifndef MaxStatSize
#ifdef Coexpr
#define MaxStatSize		20480	/* size of the static region in bytes*/
#else					/* Coexpr */
#define MaxStatSize		 1024	/* size of the static region in bytes */
#endif					/* Coexpr */
#endif					/* MaxStatSize */

#ifndef MaxStrSpace
#define MaxStrSpace		65000	/* size of the string space in bytes */
#endif					/* MaxStrSpace */

#ifndef MaxAbrSize
#define MaxAbrSize		65000	/* size of the block region in bytes */
#endif					/* MaxAbrSize */

#ifndef MStackSize
#define MStackSize		10000	/* size of the main stack in words */
#endif					/* MStackSize */

#ifndef StackSize
#define StackSize		 2000	/* words in co-expression stack */
#endif					/* StackSize */

#ifndef QualLstSize
#define QualLstSize		 5000	/* size of qualifier pointer region */
#endif					/* QualLstSize */

#ifndef ActStkBlkEnts
#ifdef Coexpr
#define ActStkBlkEnts		  100	/* number of entries in an astkblk */
#else					/* Coexpr */
#define ActStkBlkEnts		    1	/* number of entries in an astkblk */
#endif					/* Coexpr */
#endif					/* ActStkBlkEnts */

/*
 * Minimum regions sizes (presently not used).
 */

#ifndef MinStatSize
#ifdef Coexpr
#define MinStatSize		10240	/* size of the static region in bytes*/
#else					/* Coexpr */
#define MinStatSize		 1024	/* size of static region in bytes */
#endif					/* Coexpr */
#endif					/* MinStatSize */

#ifndef MinStrSpace
#define MinStrSpace		 5000	/* size of the string space in bytes */
#endif					/* MinStrSpace */

#ifndef MinAbrSize
#define MinAbrSize		 5000	/* size of the block region in bytes */
#endif					/* MinAbrSize */

#ifndef MinMStackSize
#define MinMStackSize		 2000	/* size of the main stack in words */
#endif					/* MinMStackSize */

#ifndef MinStackSize
#define MinStackSize		 1000	/* words in co-expression stack */
#endif					/* MinStackSize */

#ifndef MinQualLstSize
#define MinQualLstSize		  500	/* size of qualifier pointer region */
#endif					/* MinQualLstSize */

#ifndef GranSize
#define GranSize	     	   64	/* storage allocation granule size */
#endif					/* GranSize */

#ifndef Sqlinc
#define Sqlinc	    128*sizeof(dptr *) 	/* qualifier pointer list increment */
#endif					/* Sqlinc */

/*
 * Codes returned by runtime support routines.
 *  Note, some conversion routines also return type codes. Other routines may
 *  return positive values other than return codes. sort() places restrictions
 *  on Less, Equal, and Greater.
 */

#define Less		-1
#define Equal		0
#define Greater		1
#define CvtFail		-2
#define Cvt		-3
#define NoCvt		-4
#define Failure		-5
#define Defaulted	-6
#define Success		-7
#define Error		-8

#undef ToAscii
#undef FromAscii
#if EBCDIC == 2
#define ToAscii(e) (FromEBCDIC[e])
#define FromAscii(e) (ToEBCDIC[e])
#else					/* EBCDIC == 2 */
#define ToAscii(e) (e)
#define FromAscii(e) (e)
#endif					/* EBCDIC == 2 */

/*
 * Pointer to block.
 */
#define BlkLoc(d)	((d).vword.bptr)

/*
 * Check for null-valued descriptor.
 */
#define ChkNull(d)	((d).dword==D_Null)

/*
 * Check for equivalent descriptors.
 */
#define EqlDesc(d1,d2)	((d1).dword == (d2).dword && BlkLoc(d1) == BlkLoc(d2))

/*
 * Integer value.
 */
#define IntVal(d)	((d).vword.integr)

/*
 * Offset from top of block to value of variable.
 */
#define Offset(d)	((d).dword & OffsetMask)

/*
 * Check for pointer.
 */
#define Pointer(d)	((d).dword & F_Ptr)

/*
 * Check for qualifier.
 */
#define Qual(d)		(!((d).dword & F_Nqual))

/*
 * Length of string.
 */
#define StrLen(q)	((q).dword)

/*
 * Location of first character of string.
 */
#define StrLoc(q)	((q).vword.sptr)
/*
 * Type of descriptor.
 */
#define Type(d)		(int)((d).dword & TypeMask)

/*
 * Check for variable.
 */
#define Var(d)		((d).dword & F_Var)

/*
 * Location of the value of a variable.
 */
#define VarLoc(d)	((d).vword.descptr)

/*
 *  Important note:  The code that follows is not strictly legal C.
 *   It tests to see if pointer p2 is between p1 and p3. This may
 *   involve the comparison of pointers in different arrays, which
 *   is not well-defined.  The casts of these pointers to unsigned "words"
 *   (longs or ints, depending) works with all C compilers and architectures
 *   on which Icon has been implemented.  However, it is possible it will
 *   not work on some system.  If it doesn't, there may be a "false
 *   positive" test, which is likely to cause a memory violation or a
 *   loop. It is not practical to implement Icon on a system on which this
 *   happens.
 */

#define InRange(p1,p2,p3) ((uword)(p2) >= (uword)(p1) && (uword)(p2) < (uword)(p3))

/*
 * Get floating-point number from real block.
 */
#ifdef Double
#define GetReal(dp,res)	{ \
                         word *rp, *rq; \
                         rp = (word *) &(res); \
                         rq = (word *) &(BlkLoc(*dp)->realblk.realval); \
                         *rp++ = *rq++; \
                         *rp = *rq;} 
#else					/* Double */
#define GetReal(dp,res)	res = BlkLoc(*dp)->realblk.realval
#endif					/* Double */

/*
 * Absolute value of x (word).
 */
#if SASC
#define Abs(x) __builtin_abs(x)
#else					/* SASC */
#define Abs(x) (((x) < 0) ? (-(x)) : (x))
#endif					/* SASC */

/*
 * Maximum of x and y.
 */
#define Max(x,y)        ((x)>(y)?(x):(y))
#if SASC
#undef Max
#define Max(x,y)     __builtin_max(x,y)
#endif					/* SASC */

/*
 * Minimum of x and y.
 */
#define Min(x,y)        ((x)<(y)?(x):(y))
#if SASC
#undef Min
#define Min(x,y)     __builtin_min(x,y)
#endif					/* SASC */

/*
 * Some C compilers take '\n' and '\r' to be the same, so the
 *  following definitions are used.
 */
#if EBCDIC
/*
 * Note that, in EBCDIC, "line feed" and "new line" are distinct
 *  characters.  Icon's use of "line feed" is really "new line" in
 *  C terms.
 */
#define LineFeed '\n' /* if really "line feed", that's 37 */
#define CarriageReturn '\r'
#else					/* EBCDIC */
#define LineFeed  10
#define CarriageReturn 13
#endif					/* EBCDIC */

/*
 * Construct an integer descriptor.
 */
#define MakeInt(i,dp)	{ \
                 	 (dp)->dword = D_Integer; \
                         IntVal(*dp) = (word)(i);}

/*
 * Offset in word of cset bit.
 */
#define CsetOff(b)	((b) & BitOffMask) 
/*
 * Set bit b in cset c.
 */
#define Setb(b,c)	(*CsetPtr(b,c) |= (01 << CsetOff(b))) 
/*
 * Test bit b in cset c.
 */
#define Testb(b,c)	((*CsetPtr(b,c) >> CsetOff(b)) & 01) 

/*
 *  Vsizeof is for use with variable-sized (i.e., indefinite)
 *   structures containing an array of descriptors declared of size 1
 *   to avoid compiler warnings associated with 0-sized arrays.
 */

#define Vsizeof(s)	(sizeof(s) - sizeof(struct descrip))

/*
 * Other sizeof macros:
 *
 *  Wsizeof(x)	-- Size of x in words.
 *  Vwsizeof(x) -- Size of x in words, minus the size of a descriptor.	Used
 *   when structures have a potentially null list of descriptors
 *   at their end.
 */
#define Wsizeof(x)	((sizeof(x) + sizeof(word) - 1) / sizeof(word))
#define Vwsizeof(x)	((sizeof(x) - sizeof(struct descrip) +sizeof(word) - 1)\
			   / sizeof(word))
/*
 * Definitions and declarations used for storage management.
 */

#define F_Mark		0100000 	/* bit for marking blocks */

#define Static  1			/* collection is for static region */
#define Strings	2			/* collection is for strings */
#define Blocks	3			/* collection is for blocks */

/*
 * Get type of block pointed at by x.
 */
#define BlkType(x)   (*(word *)x)

/*
 * BlkSize(x) takes the block pointed to by x and if the size of
 *  the block as indicated by bsizes[] is nonzero it returns the
 *  indicated size; otherwise it returns the second word in the
 *  block contains the size.
 */
#define BlkSize(x) (bsizes[*(word *)x & ~F_Mark] ? \
		     bsizes[*(word *)x & ~F_Mark] : *((word *)x + 1))


#if COMPILER
/*
 * Definitions for the compiler.
 */

/*
 * If memory monitoring is not enabled, redefine function calls
 * to do nothing.  *** this needs to be merged with monitor.h ****
 */
#ifndef MemMon
#define MMAlc(n,t)
#define MMBGC(r)
#define MMEGC()
#define MMMark(b,t)
#define MMShow(dp,s)
#define MMStat(a,l,c)
#define MMStr(n)
#define MMSMark(a,n)
#endif					/* MemMon */

#define k_pos kywd_pos.vword.integr	/* value of &pos */
#define k_random kywd_ran.vword.integr	/* value of &random */
#define k_trace kywd_trc.vword.integr	/* value of &trace */

/*
 * Type codes (descriptors and blocks).
 */
#define T_Null		 0	/* null value */
#define T_Integer	 1	/* integer */
#define T_Real		 3	/* real number */
#define T_Cset		 4	/* cset */
#define T_File		 5	/* file */
#define T_Proc		 6	/* procedure */
#define T_List		 7	/* list header */
#define T_Table		 8	/* table header */
#define T_Record	 9	/* record */
#define T_Telem		10	/* table element */
#define T_Lelem		11	/* list element */
#define T_Tvsubs	12	/* substring trapped variable */
#define T_Tvtbl		13	/* table element trapped variable */
#define T_Kywdint	14	/* integer keyword */
#define T_Kywdpos	15	/* keyword &pos */
#define T_Kywdsubj	16	/* keyword &subject */
#define T_Set		17	/* set header */
#define T_Selem		18	/* set element */
#define T_Refresh	19	/* refresh block */
#define T_Coexpr	20	/* co-expression */
#define T_External	21	/* external block */

#define MaxType		21	/* maximum type number */

/*
 * Descriptor types and flags.
 */

#define D_Null		(T_Null     | D_Typecode)
#define D_Integer	(T_Integer  | D_Typecode)
#define D_Real		(T_Real     | D_Typecode | F_Ptr)
#define D_Cset		(T_Cset     | D_Typecode | F_Ptr)
#define D_File		(T_File     | D_Typecode | F_Ptr)
#define D_Proc		(T_Proc     | D_Typecode | F_Ptr)
#define D_List		(T_List     | D_Typecode | F_Ptr)
#define D_Table		(T_Table    | D_Typecode | F_Ptr)
#define D_Telem		(T_Telem    | D_Typecode | F_Ptr)
#define D_Set		(T_Set      | D_Typecode | F_Ptr)
#define D_Record	(T_Record   | D_Typecode | F_Ptr)
#define D_Tvsubs	(T_Tvsubs   | D_Typecode | F_Ptr | F_Var)
#define D_Tvtbl		(T_Tvtbl    | D_Typecode | F_Ptr | F_Var)
#define D_Kywdint	(T_Kywdint  | D_Typecode | F_Ptr | F_Var)
#define D_Kywdpos	(T_Kywdpos  | D_Typecode | F_Ptr | F_Var)
#define D_Kywdsubj	(T_Kywdsubj | D_Typecode | F_Ptr | F_Var)
#define D_Refresh	(T_Refresh  | D_Typecode | F_Ptr)
#define D_Coexpr	(T_Coexpr   | D_Typecode | F_Ptr)
#define D_External	(T_External | D_Typecode | F_Ptr)

#define D_Var		(F_Var | F_Nqual | F_Ptr)
#define D_Typecode	(F_Nqual | F_Typecode)

#define TypeMask	63	/* type mask */
#define OffsetMask	(~(D_Var)) /* offset mask for variables */

#define MaxCvtLen	257	/* largest string in conversions */

#define E_Resume	-1	/* expression failed: resume a generator */
#define E_Continue      -2	/* expression returned: continue execution */
#define E_FallThru      -3      /* body function: fell through end of code */

#define A_Coact		1	/* co-expression ativation */
#define A_Coret		2	/* co-expression return */
#define A_Cofail	3	/* co-expression failure */

#define SSlots		   37	/* number of slots in a set block */
#define TSlots		   37	/* number of slots in a table block */

/*
 * "In place" dereferencing.
 */
#define Deref(d) if (Var(d)) deref(d, d);

/*
 * Construct a substring trapped variable.
 */
#define SubStr(dest,var,len,pos)\
   if ((var)->dword == D_Tvsubs)\
      (dest)->vword.bptr = (union block *)alcsubs(len, (pos) +\
         BlkLoc(*(var))->tvsubs.sspos - 1, &BlkLoc(*(var))->tvsubs.ssvar);\
   else\
      (dest)->vword.bptr = (union block *)alcsubs(len, pos, (var));\
   (dest)->dword = D_Tvsubs;

/*
 * Get slot number in set or table.
 */
#if IntBits == 16
#define SlotNum(i,j)	((word)(i&~(1<<(WordBits - 1))) % j)
#else					/* IntBits == 16 */
#define SlotNum(i,j)	((i&~(1<<(WordBits - 1))) % j)
#endif					/* IntBits == 16 */

/*
 * Address of word containing cset bit b (c is a struct descrip of type Cset).
 */
#define CsetPtr(b,c)	(BlkLoc(c)->cset.bits + (((b)&0377) >> LogIntBits)) 

/*
 * Find debug struct in procedure frame, assuming debugging is enabled.
 *  Note that there is always one descriptor in array even if it is not
 *  being used.
 */
#define PFDebug(pf) ((struct debug *)((char *)(pf).tend.d +\
    sizeof(struct descrip) * ((pf).tend.num ? (pf).tend.num : 1)))

/*
 * Handy sizeof macros:
 *
 *  Wsizeof(x)	-- Size of x in words.
 *  Vwsizeof(x) -- Size of x in words, minus the size of a descriptor.	Used
 *   when structures have a potentially null list of descriptors
 *   at their end.
 */
#define Wsizeof(x)	((sizeof(x) + sizeof(word) - 1) / sizeof(word))
#define Vwsizeof(x)	((sizeof(x) - sizeof(struct descrip) +sizeof(word) - 1)\
			   / sizeof(word))

/*
 * Macro for initialized procedure block.
 */
#define B_IProc(n) struct {word title; word blksize; int (*ccode)();\
   word nparam; word ndynam; word nstatic; word fstatic;\
   struct sdescrip quals[n];} 
#else					/* COMPILER */

/*
 * Definitions for the interpreter.
 */

/*
 * Definitions for interpreter actions.
 */
#define A_Failure	1		/* routine failed */
#define A_Suspension	2		/* routine suspended */
#define A_Return	3		/* routine returned */
#define A_Pret_uw	4		/* interp unwind for Op_Pret */
#define A_Unmark_uw	5		/* interp unwind for Op_Unmark */
#define A_Resumption	6		/* resume generator */
#define A_Pfail_uw	7		/* interp unwind for Op_Pfail */
#define A_Lsusp_uw	8		/* interp unwind for Op_Lsusp */
#define A_Eret_uw	9		/* interp unwind for Op_Eret */
#define A_Coact		10		/* co-expression activated */
#define A_Coret		11		/* co-expression returned */
#define A_Cofail	12		/* co-expression failed */

/*
 * Codes returned by invoke to indicate action.
 */
#define I_Builtin	201	/* A built-in routine is to be invoked */
#define I_Fail		202	/* goal-directed evaluation failed */
#define I_Continue	203	/* Continue execution in the interp loop */
#define I_Vararg	204	/* A function with a variable number of args */

/*
 * Generator types.
 */
#define G_Csusp		1
#define G_Esusp		2
#define G_Psusp		3

/*
 * Type codes (descriptors and blocks).
 */
#define	T_String	-1	/* string -- for reference; not used */
#define T_Null		 0	/* null value */
#define T_Integer	 1	/* integer */

#ifdef LargeInts
#define T_Lrgint	 2	/* long integer */
#endif					/* LargeInts */

#define T_Real		 3	/* real number */
#define T_Cset		 4	/* cset */
#define T_File		 5	/* file */
#define T_Proc		 6	/* procedure */
#define T_List		 7	/* list header */
#define T_Table		 8	/* table header */
#define T_Record	 9	/* record */
#define T_Telem		10	/* table element */
#define T_Lelem		11	/* list element */
#define T_Tvsubs	12	/* substring trapped variable */
#define T_Tvkywd	13	/* keyword trapped variable */
#define T_Tvtbl		14	/* table element trapped variable */
#define T_Set		15	/* set header */
#define T_Selem		16	/* set element */
#define T_Refresh	17	/* refresh block */
#define T_Coexpr	18	/* co-expression */
#define T_External	19	/* external block */
#define T_Slots		20	/* set/table hash slots */

#define MaxType		20	/* maximum type number */

/*
 * Descriptor types and flags.
 */

#define D_Null		(word)(T_Null | F_Nqual)
#define D_Integer	(word)(T_Integer | F_Nqual)

#ifdef LargeInts
#define D_Lrgint	(word)(T_Lrgint | F_Ptr | F_Nqual)
#endif					/* LargeInts */

#define D_Real		(word)(T_Real | F_Ptr | F_Nqual)
#define D_Cset		(word)(T_Cset | F_Ptr | F_Nqual)
#define D_File		(word)(T_File | F_Ptr | F_Nqual)
#define D_Proc		(word)(T_Proc | F_Ptr | F_Nqual)
#define D_List		(word)(T_List | F_Ptr | F_Nqual)
#define D_Table		(word)(T_Table | F_Ptr | F_Nqual)
#define D_Telem		(word)(T_Telem | F_Ptr | F_Nqual)
#define D_Tvsubs	(word)(T_Tvsubs | D_Tvar)
#define D_Tvkywd	(word)(T_Tvkywd | D_Tvar)
#define D_Tvtbl		(word)(T_Tvtbl | D_Tvar)
#define D_Record	(word)(T_Record | F_Ptr | F_Nqual)
#define D_Set		(word)(T_Set | F_Ptr | F_Nqual)
#define D_Refresh	(word)(T_Refresh | F_Ptr | F_Nqual)
#define D_Coexpr	(word)(T_Coexpr | F_Ptr | F_Nqual)
#define D_External	(word)(T_External | F_Ptr | F_Nqual)
#define D_Slots		(word)(T_Slots | F_Ptr | F_Nqual)

#define D_Var		(word)(F_Var | F_Nqual | F_Ptr)
#define D_Tvar		(word)(D_Var | F_Tvar)

#define TypeMask	63	/* type mask */
#define OffsetMask	(~(D_Tvar)) /* offset mask for variables */


#define k_error tvky_err.kyval.vword.integr	/* value of &error */
#define k_pos tvky_pos.kyval.vword.integr	/* value of &pos */
#define k_random tvky_ran.kyval.vword.integr	/* value of &random */
#define k_subject tvky_sub.kyval		/* value of &subject */
#define k_trace tvky_trc.kyval.vword.integr	/* value of &trace */

/*
 * Evaluation stack overflow margin
 */

#define PerilDelta 100

/*
 * Macro definitions related to descriptors.
 */

/*
 * The following code is operating-system dependent [@rt.01].  Define
 *  PushAval for computers that store longs and pointers differently.
 */

#if PORT
#define PushAVal(x) PushVal(x)
Deliberate Syntax Error
#endif					/* PORT */

#if AMIGA || ARM || ATARI_ST || HIGHC_386 || MACINTOSH || MVS || UNIX || VM || VMS
#define PushAVal(x) PushVal(x)
#endif					/* AMIGA || ARM || ATARI_ST ... */

#if MSDOS || OS2
static union {
       pointer stkadr;
       word stkint;
   } stkword;

#define PushAVal(x)  {sp++; \
			stkword.stkadr = (char *)(x); \
			*sp = stkword.stkint;}
#endif					/* MSDOS || OS2 */

/*
 * End of operating-system specific code.
 */

/*
 * Dereference descriptor.
 */
#define DeRef(d)	(Var(d) ? deref(&d) : Success)

/*
 * Check for trapped variable.
 */
#define Tvar(d)		((d).dword & F_Tvar)

/*
 * Location of trapped-variable block.
 */
#define TvarLoc(d)	((d).vword.bptr)

/*
 * Macros for pushing values on the interpreter stack.
 */

/*
 * Push descriptor.
 */
#define PushDesc(d)	{*++sp = ((d).dword); sp++;*sp =((d).vword.integr);}

/*
 * Push null-valued descriptor.
 */
#define PushNull	{*++sp = D_Null; sp++; *sp = 0;}

/*
 * Push word.
 */
#define PushVal(v)	{*++sp = (word)(v);}

/*
 * Macros related to function and operator definition.
 */

/*
 * Procedure block for a function.
 */


#define FncBlock(f,nargs,deref) \
	struct b_iproc Cat(B,f) = {\
	T_Proc,\
	Vsizeof(struct b_proc),\
	Cat(X,f),\
	nargs,\
	-1,\
	deref, 0,\
	{sizeof(Lit(f))-1,Lit(f)}};


/*
 * Function declaration for variable number of arguments.
 */
#define FncDcl(nm,n) FncBlock(nm,n,0) Cat(X,nm)(cargp)  register dptr cargp;

/*
 * Function declaration for variable number of arguments.
 */
#define FncDclV(nm) FncBlock(nm,-1,0) Cat(X,nm)(nargs,cargp) int nargs; register dptr cargp;

/*
 * Function declaration without dereferenced arguments.
 */
#define FncNDcl(nm,n) FncBlock(nm,n,-1) Cat(X,nm)(cargp)  register dptr cargp;

/*
 * Function declaration for variable number of arguments.
 */
#define FncNDclV(nm) FncBlock(nm,-1,-1) Cat(X,nm)(nargs,cargp) int nargs; register dptr cargp;

/*
 * Declaration for library routine.
 */
#define LibDcl(nm,n,pn) OpBlock(nm,n,pn,0) Cat(O,nm)(nargs,cargp) \
   register dptr cargp;
/*
 * Procedure block for an operator.
 */
#define OpBlock(f,nargs,sname,xtrargs)\
	struct b_iproc Cat(B,f) = {\
	T_Proc,\
	Vsizeof(struct b_proc),\
	Cat(O,f),\
	nargs,\
	-1,\
	xtrargs,\
	0,\
	{sizeof(sname)-1,sname}};

/*
 * Operator declaration.
 */
#define OpDcl(nm,n,pn) OpBlock(nm,n,pn,0) Cat(O,nm)(cargp) register dptr cargp;

/*
 * Operator declaration with extra working argument.
 */
#define OpDclE(nm,n,pn) OpBlock(nm,-n,pn,0) Cat(O,nm)(cargp) register dptr cargp;

/*
 * Agent routine declaration.
 */
#define AgtDcl(nm) Cat(A,nm)(cargp) register dptr cargp;

#ifdef StrInvoke
/*
 * Structure for mapping string names of procedures to block addresses.
 */
struct pstrnm {
   char *pstrep;
   struct b_proc *pblock;
   };

#endif					/* StrInvoke */

/*
 * Macros to access Icon arguments in C functions.
 */

/*
 * n-th argument.
 */
#define Arg(n)	 	(cargp[n])

/*
 * Type field of n-th argument.
 */
#define ArgType(n)	(cargp[n].dword)

/*
 * Value field of n-th argument.
 */
#define ArgVal(n)	(cargp[n].vword.integr)

/*
 * Specific arguments.
 */
#define Arg0	(cargp[0])
#define Arg1	(cargp[1])
#define Arg2	(cargp[2])
#define Arg3	(cargp[3])
#define Arg4	(cargp[4])
#define Arg5	(cargp[5])
#define Arg6	(cargp[6])
#define Arg7	(cargp[7])
#define Arg8	(cargp[8])

/*
 * Code expansions for exits from C code for top-level routines.
 */
#define Fail		return A_Failure
#define Return		return A_Return

#define Suspend  { \
   int rc; \
   if ((rc = interp(G_Csusp,cargp)) != A_Resumption) \
      return rc;} 

#define Forward(agent) return Cat(A,agent)(cargp)

/*
 * Miscellaneous macro definitions.
 */

/*
 * Error exit from non top-level routines.
 */
#define RetError(n,v) {\
   k_errornumber = n;\
   k_errortext = "";\
   k_errorvalue = v;\
   return Error;}

/*
 * Check whether a set or table needs resizing.
 */
#define SP(p) ((struct b_set *)p)
#define TooCrowded(p) \
   ((SP(p)->size > MaxHLoad*(SP(p)->mask+1)) && (SP(p)->hdir[HSegs-1] == NULL))
#define TooSparse(p) \
   ((SP(p)->hdir[1] != NULL) && (SP(p)->size < MinHLoad*(SP(p)->mask+1)))

/*
 * RunErr encapsulates a call to the function runerr, followed
 *  by Fail.  The idea is to avoid the problem of calling
 *  runerr directly and forgetting that it may actually return.
 */

#define RunErr(n,dp) {\
   runerr((int)n,dp);\
   Fail;\
   }

/*
 * Address of word of cset bit.
 */
#define CsetPtr(b,c)	((c) + (((b)&0377) >> LogIntBits)) 


#define StackBase(ce) BlkLoc(ce)
#endif					/* COMPILER */
