
/*
 * fmath.c -- sin, cos, tan, acos, asin, atan, dtor, rtod, exp, log, sqrt
 */

#include "../h/rt.h"

#ifdef MathFncs

#ifndef PI
#define PI 3.14159
#endif					/* PI */

/*
 * sin(x), x in radians
 */

FncDcl(sin,1)
   {
   int t;
   double sin();

   if ((t = cvreal(&Arg1)) == CvtFail) 
     RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(sin(valbuffer), &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
   if (makereal(sin(BlkLoc(Arg1)->realblk.realval), &Arg0) == Error) 
      RunErr(0, NULL);
#endif					/* Double */

   Return;
   }

/*
 * cos(x), x in radians
 */

FncDcl(cos,1)
   {
   int t;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(cos(valbuffer), &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
   if (makereal(cos(BlkLoc(Arg1)->realblk.realval), &Arg0) == Error) 
      RunErr(0, NULL);
#endif					/* Double */

   Return;
   }

/*
 * tan(x), x in radians
 */

FncDcl(tan,1)
   {
   int t;
   double y;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);
   errno = 0;

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		y = tan(valbuffer);
	}
#else					/* Double */
   y = tan(BlkLoc(Arg1)->realblk.realval);
#endif					/* Double */

   if (errno == ERANGE) 
      RunErr(-204, NULL);
   if (makereal(y, &Arg0) == Error) 
      RunErr(0, NULL);
   Return;
   }

/*
 * acos(x), x in radians
 */
FncDcl(acos,1)
   {
   int t;
   double r, y;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &r;
		*rq++ = *rp++;
		*rq = *rp;
	}
#else					/* Double */
   r = BlkLoc(Arg1)->realblk.realval;
#endif					/* Double */

   if (r < -1.0 || r > 1.0)		/* can't count on library */
      RunErr(205,&Arg1);
   errno = 0;
   y = acos(r);
   if (errno == EDOM) 
      RunErr(-205, NULL);
   if (makereal(y, &Arg0) == Error) 
      RunErr(0, NULL);
   Return;
   }

/*
 * asin(x), x in radians
 */
FncDcl(asin,1)
   {
   int t;
   double r, y;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &r;
		*rq++ = *rp++;
		*rq = *rp;
	}
#else					/* Double */
   r = BlkLoc(Arg1)->realblk.realval;
#endif					/* Double */

   if (r < -1.0 || r > 1.0)		/* can't count on library */
      RunErr(205,&Arg1);
   errno = 0;
   y = asin(r);
   if (errno == EDOM) 
      RunErr(-205, NULL);
   if (makereal(y, &Arg0) == Error) 
      RunErr(0, NULL);
   Return;
   }

/*
 * atan(x,y) -- x,y  in radians; if y is present, produces atan2(x,y).
 */
FncDcl(atan,2)
   {
   int t;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);
   if (ChkNull(Arg2)) {

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(atan(valbuffer), &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
      if (makereal(atan(BlkLoc(Arg1)->realblk.realval), &Arg0) == Error) 
         RunErr(0, NULL);
#endif					/* Double */

      }
   else {
      if ((t = cvreal(&Arg2)) == CvtFail) 
         RunErr(102, &Arg2);

#if AMIGA
#if AZTEC_C
#define atan2(x,y) atan(x/y)
#endif					/* AZTEC_C */
#endif					/* AMIGA */

#ifdef Double
	{	int	*rp, *rq;
		double	vbuffer1, vbuffer2;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &vbuffer1;
		*rq++ = *rp++;
		*rq = *rp;
		rp = (word *) &(BlkLoc(Arg2)->realblk.realval);
		rq = (word *) &vbuffer2;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(atan2(vbuffer1, vbuffer2), &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
      if (makereal(atan2(BlkLoc(Arg1)->realblk.realval,
		       BlkLoc(Arg2)->realblk.realval), &Arg0) == Error) 
         RunErr(0, NULL);
#endif					/* Double */

      }
   Return;
   }

/*
 * dtor(x), x in degrees
 */

FncDcl(dtor,1)
   {

   if (cvreal(&Arg1) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(valbuffer * PI / 180, &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
   if (makereal(BlkLoc(Arg1)->realblk.realval * PI / 180, &Arg0) == Error) 
      RunErr(0, NULL);
#endif					/* Double */

   Return;
   }

/*
 * rtod(x), x in radians
 */
FncDcl(rtod,1)
   {

   if (cvreal(&Arg1) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (makereal(valbuffer * 180 / PI, &Arg0) == Error)
			RunErr(0, NULL);
	}
#else					/* Double */
   if (makereal(BlkLoc(Arg1)->realblk.realval * 180 / PI, &Arg0) == Error) 
      RunErr(0, NULL);
#endif					/* Double */

   Return;
   }

/*
 * exp(x)
 */

FncDcl(exp,1)
   {
   int t;
   double y;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);
   errno = 0;

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		y = exp(valbuffer);
	}
#else					/* Double */
   y = exp(BlkLoc(Arg1)->realblk.realval);
#endif					/* Double */
   if (errno == ERANGE) 
      RunErr(-204, NULL);
   if (makereal(y, &Arg0) == Error) 
      RunErr(0, NULL);
   Return;
   }

/*
 * log(x,b) - logarithm of x to base b.
 */
FncDcl(log,2)
   {
   static double lastbase = 0.0;
   static double divisor;
   double x;
 
   if (cvreal(&Arg1) != T_Real)
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (valbuffer <= 0.0)
			RunErr(205, &Arg1);
	}
#else					/* Double */
   if (BlkLoc(Arg1)->realblk.realval <= 0.0)
      RunErr(205, &Arg1);
#endif					/* Double */

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		x = log(valbuffer);
	}
#else					/* Double */
   x = log(BlkLoc(Arg1)->realblk.realval);
#endif					/* Double */

   if (! ChkNull(Arg2))  {
      if (cvreal(&Arg2) != T_Real)
         RunErr(102, &Arg2);

#ifdef Double
	{	int	*rp, *rq;
		double	valbuffer;
		rp = (word *) &(BlkLoc(Arg2)->realblk.realval);
		rq = (word *) &valbuffer;
		*rq++ = *rp++;
		*rq = *rp;
		if (valbuffer <= 1.0)
			RunErr(205, &Arg2);
		if (valbuffer != lastbase) {
			divisor = log(valbuffer);
			lastbase = valbuffer;
		}
	}
#else					/* Double */
      if (BlkLoc(Arg2)->realblk.realval <= 1.0)
         RunErr(205, &Arg2);
      if (BlkLoc(Arg2)->realblk.realval != lastbase) {
         divisor = log(BlkLoc(Arg2)->realblk.realval);
         lastbase = BlkLoc(Arg2)->realblk.realval;
         }
#endif					/* Double */

      x = x / divisor;
      }  
   if (makereal(x, &Arg0) == Error)
      RunErr(0, NULL);
   Return;
   }


/*
 * sqrt(x)
 */

FncDcl(sqrt,1)
   {
   int t;
   double r, y;

   if ((t = cvreal(&Arg1)) == CvtFail) 
      RunErr(102, &Arg1);

#ifdef Double
	{	int	*rp, *rq;
		rp = (word *) &(BlkLoc(Arg1)->realblk.realval);
		rq = (word *) &r;
		*rq++ = *rp++;
		*rq = *rp;
	}
#else					/* Double */
   r = BlkLoc(Arg1)->realblk.realval;
#endif					/* Double */

   if (r < 0)
      RunErr(205, &Arg1);
   y = sqrt(r);
   errno = 0;
   if (errno == EDOM) 
      RunErr(-205, NULL);
   if (makereal(y, &Arg0) == Error) 
      RunErr(0, NULL);
   Return;
   }
#else					/* MathFncs */
static char x;			/* prevent empty module */
#endif					/* MathFncs */
