/*
 * File: ralc.c
 *  Contents: allocation routines.
 */

#include "../h/rt.h"

/*
 * Prototype.
 */

hidden	union   block *alcblk   Params((uword nbytes,int tcode));

extern int qualfail;
extern word coll_blk;
extern word coll_str;
extern word alcnum;

word coexp_ser = 2;	/* serial numbers for co-expressions; &main is 1 */
word list_ser = 1;	/* serial numbers for lists */
word set_ser = 1;	/* serial numbers for sets */
word table_ser = 1;	/* serial numbers for tables */

/*
 * alcblk - returns pointer to nbytes of free storage in block region.
 */

static union block *alcblk(nbytes,tcode)
uword nbytes;
int tcode;
   {
   register uword fspace, *sloc;

   /*
    * See if there is enough room in the block region.
    */
   fspace = DiffPtrs(blkend,blkfree);
   if (fspace < nbytes)
      syserr("block allocation botch");

   /*
    * If monitoring, show the allocation.
    */
   MMAlc((word)nbytes,tcode);

   /*
    * Decrement the free space in the block region by the number of bytes
    *  allocated and return the address of the first byte of the allocated
    *  block.
    */
   sloc = (uword *)blkfree;
   blkneed -= nbytes;
   blkfree += nbytes;
   BlkType(sloc) = tcode;
   return (union block *)(sloc);
   }

/*
 * alcreal - allocate a real value in the block region.
 */

struct b_real *alcreal(val)
double val;
   {
   register struct b_real *blk;

   blk = (struct b_real *)alcblk((uword)sizeof(struct b_real), T_Real);

#ifdef Double
/* access real values one word at a time */
   { int *rp, *rq;
     rp = (word *) &(blk->realval);
     rq = (word *) &val;
     *rp++ = *rq++;
     *rp   = *rq;
   }
#else                                   /* Double */
   blk->realval = val;
#endif                                  /* Double */

   return blk;
   }

#ifdef LargeInts
/*
 * alcbignum - allocate an n-digit bignum in the block region
 */

struct b_bignum *alcbignum(n)
word n;
   {
   register struct b_bignum *blk;
   register uword size;

   size = sizeof(struct b_bignum) + ((n - 1) * sizeof(DIGIT));
   /* ensure whole number of words allocated */
   size = (size + WordSize - 1) & -WordSize;
   blk = (struct b_bignum *)alcblk(size, T_Lrgint);
   blk->blksize = size;
   blk->msd = blk->sign = 0;
   blk->lsd = n - 1;
   return blk;
   }
#endif					/* LargeInts */

/*
 * alccset - allocate a cset in the block region.
 */

struct b_cset *alccset()
   {
   register struct b_cset *blk;
   register int i;

   blk = (struct b_cset *)alcblk((uword)sizeof(struct b_cset), T_Cset);
   blk->size = -1;              /* flag size as not yet computed */

   /*
    * Zero the bit array.
    */
   for (i = 0; i < CsetSize; i++)
     blk->bits[i] = 0;
   return blk;
   }

/*
 * alcfile - allocate a file block in the block region.
 */

struct b_file *alcfile(fd, status, name)
FILE *fd;
int status;
dptr name;
   {
   register struct b_file *blk;

   blk = (struct b_file *)alcblk((uword)sizeof(struct b_file), T_File);
   blk->fd = fd;
   blk->status = status;
   blk->fname = *name;
   return blk;
   }

/*
 * alcrecd - allocate record with nflds fields in the block region.
 */

struct b_record *alcrecd(nflds, recptr)
int nflds;
union block **recptr;
   {
   register struct b_record *blk;
   register int size;

   size = Vsizeof(struct b_record) + nflds*sizeof(struct descrip);
   blk = (struct b_record *)alcblk((uword)size, T_Record);
   blk->blksize = size;
   blk->recdesc = (union block *)recptr;
   return blk;
   }

/*
 * alcextrnl - allocate an external block.
 */

struct b_external *alcextrnl(n)
uword n;
   {
   register struct b_external *blk;
   register uword size;
   
   size = n * sizeof(word) + (sizeof(struct b_external) - sizeof(word));
   blk = (struct b_external *)alcblk(size, T_External);
   blk->blksize = size;
   while (n-- > 0)
      blk->exdata[n] = (word)0;		/* clear block */
   return blk;
   }

/*
 * alclist - allocate a list header block in the block region.
 */

struct b_list *alclist(size)
uword size;
   {
   register struct b_list *blk;

   blk = (struct b_list *)alcblk((uword)sizeof(struct b_list), T_List);
   blk->size = size;
   blk->listhead = NULL;
   blk->listtail = NULL;
   blk->id = list_ser++;
   return blk;
   }

/*
 * alclstb - allocate a list element block in the block region.
 */

struct b_lelem *alclstb(nslots, first, nused)
uword nslots, first, nused;
   {
   register struct b_lelem *blk;
   register word i, size;

   size = Vsizeof(struct b_lelem) + nslots * sizeof(struct descrip);
   blk = (struct b_lelem *)alcblk((uword)size, T_Lelem);
   blk->blksize = size;
   blk->nslots = nslots;
   blk->first = first;
   blk->nused = nused;
   blk->listprev = NULL;
   blk->listnext = NULL;
   /*
    * Set all elements to &null.
    */
   for (i = 0; i < nslots; i++)
      blk->lslots[i] = nulldesc;
   return blk;
   }

/*
 * alchash - allocate a hashed structure (set or table header) in the block
 *  region.
 */

union block *alchash(tcode)
int tcode;
   {

   register int i;
   register union block *blk;
   word serial;
   uword blksize;

   if (tcode == T_Table) {
      serial = table_ser++;
      blksize = sizeof(struct b_table);
      }
   else {	/* tcode == T_Set */
      serial = set_ser++;
      blksize = sizeof(struct b_set);
      }
   blk = alcblk(blksize, tcode);
   blk->set.size = 0;
   blk->set.id = serial;
   blk->set.mask = 0;
   for (i = 0; i < HSegs; i++)
      blk->set.hdir[i] = NULL;
   return blk;
   }

/*
 * alcsegment - allocate a slot block in the block region.
 */

struct b_slots *alcsegment(nslots)
word nslots;
   {
   uword size;
   register struct b_slots *blk;

   size = sizeof(struct b_slots) + WordSize * (nslots - HSlots);
   blk = (struct b_slots *)alcblk(size, T_Slots);
   blk->blksize = size;
   while (--nslots >= 0)
      blk->hslots[nslots] = NULL;
   return blk;
   }

/*
 * alctelem - allocate a table element block in the block region.
 */

struct b_telem *alctelem()
   {
   register struct b_telem *blk;

   blk = (struct b_telem *)alcblk((uword)sizeof(struct b_telem), T_Telem);
   blk->hashnum = 0;
   blk->clink = NULL;
   blk->tref = nulldesc;
   blk->tval = nulldesc;
   return blk;
   }

/*
 * alcselem - allocate a set element block.
 */

struct b_selem *alcselem(mbr,hn)
dptr mbr;
uword hn;

   {
   register struct b_selem *blk;

   blk = (struct b_selem *)alcblk((uword)sizeof(struct b_selem), T_Selem);
   blk->clink = NULL;
   blk->setmem = *mbr;
   blk->hashnum = hn;
   return blk;
   }

/*
 * alcsubs - allocate a substring trapped variable in the block region.
 */

struct b_tvsubs *alcsubs(len, pos, var)
word len, pos;
dptr var;
   {
   register struct b_tvsubs *blk;

   blk = (struct b_tvsubs *)alcblk((uword)sizeof(struct b_tvsubs), T_Tvsubs);
   blk->sslen = len;
   blk->sspos = pos;
   blk->ssvar = *var;
   return blk;
   }

/*
 * alctvtbl - allocate a table element trapped variable block in the block
 *  region.
 */

struct b_tvtbl *alctvtbl(tbl, ref, hashnum)
register dptr tbl, ref;
uword hashnum;
   {
   register struct b_tvtbl *blk;

   blk = (struct b_tvtbl *)alcblk((uword)sizeof(struct b_tvtbl), T_Tvtbl);
   blk->hashnum = hashnum;
   blk->clink = BlkLoc(*tbl);
   blk->tref = *ref;
   blk->tval = nulldesc;
   return blk;
   }

/*
 * alcstr - allocate a string in the string space.
 */

char *alcstr(s, slen)
register char *s;
register word slen;
   {
   register char *d;
   register uword fspace;
   char *ofree;

   MMStr(slen);

   /*
    * See if there is enough room in the string space.
    */
   fspace = DiffPtrs(strend,strfree);
   if (fspace < slen)
      syserr("string allocation botch");
   strneed -= slen;

   /*
    * Copy the string into the string space, saving a pointer to its
    *  beginning.  Note that s may be null, in which case the space
    *  is still to be allocated but nothing is to be copied into it.
    */
   ofree = d = strfree;
   if (s) {
      while (slen-- > 0)
         *d++ = *s++;
      }

   else
      d += slen;
   strfree = d;
   return ofree;
   }

/*
 * alccoexp - allocate a co-expression stack block.
 */
struct b_coexpr *alccoexp()

   {
   struct b_coexpr *ep;


#ifdef ATTM32
   ep = (struct b_coexpr *)coexp_salloc(); /* 3B2/15/4000 stack */
#else                                   /* ATTM32 */
   ep = (struct b_coexpr *)malloc((msize)stksize);
#endif                                  /* ATTM32 */

   /*
    * If malloc failed or if there have been too many co-expression allocations
    * since a collection, attempt to free some co-expression blocks and retry.
    */

#ifdef FixedRegions
   if (ep == NULL || alcnum > AlcMax) {
#else                                   /* FixedRegions */
   if (ep == NULL) {
#endif                                  /* Fixed Regions */

      collect(Static);


#ifdef ATTM32
	 ep = (struct b_coexpr *)coexp_salloc(); /* 3B2/15/4000 stack */
#else                                   /* ATTM32 */
         ep = (struct b_coexpr *)malloc((msize)stksize);
#endif                                  /* ATTM32 */
      }

   if (ep == NULL) {
      k_errornumber = -305;
      k_errortext = "";
      k_errorvalue = nulldesc;
      return NULL;
      }

#ifdef FixedRegions
   alcnum++;                    /* increment allocation count since last g.c. */
#endif                                  /* FixedRegions */

   ep->title = T_Coexpr;
   ep->es_actstk = NULL;
   ep->size = 0;
   ep->id = coexp_ser++;
   ep->nextstk = stklist;


   stklist = ep;
   MMStat((char *)ep, stksize, E_Coexpr);
   return ep;
   }

/*
 * alcactiv - allocate a co-expression activation block.
 */

struct astkblk *alcactiv()
   {
   struct astkblk *abp;

   abp = (struct astkblk *)malloc((msize)sizeof(struct astkblk));

#ifdef FixedRegions
   /*
    * If malloc failed, attempt to free some co-expression blocks and retry.
    */
   if (abp == NULL) {
      collect(Static);
      abp = (struct astkblk *)malloc((msize)sizeof(struct astkblk));
      }
#endif                                  /* FixedRegions */

   if (abp == NULL) {
      k_errornumber = -305;
      k_errortext = "";
      k_errorvalue = nulldesc;
      return NULL;
      }
   abp->nactivators = 0;
   abp->astk_nxt = NULL;
   return abp;
   }

/*
 * alcrefresh - allocate a co-expression refresh block.
 */

struct b_refresh *alcrefresh(entryx, na, nl)
word *entryx;
int na, nl;
   {
   int size;
   struct b_refresh *blk;

   size = Vsizeof(struct b_refresh) + (na + nl) * sizeof(struct descrip);
   blk = (struct b_refresh *)alcblk((uword)size, T_Refresh);
   blk->blksize = size;
   blk->ep = entryx;
   blk->numlocals = nl;
   return blk;
   }

/*
 * blkreq - insure that at least bytes of space are left in the block region.
 *  The amount of space needed is transmitted to the collector via
 *  the global variable blkneed.
 */

int blkreq(bytes)
uword bytes;
   {
   blkneed = bytes;
   if (bytes > (uword)DiffPtrs(blkend,blkfree)) {
      coll_blk++;
      collect(Blocks);
      if (bytes > (uword)DiffPtrs(blkend,blkfree))
         RetError(-307, nulldesc);
      }
   return Success;
   }

/*
 * strreq - insure that at least n of space are left in the string
 *  space.  The amount of space needed is transmitted to the collector
 *  via the global variable strneed.
 */

int strreq(n)
uword n;
   {
   strneed = n;                 /* save in case of collection */
   if (n > (uword)DiffPtrs(strend,strfree)) {
      coll_str++;
      collect(Strings);
      if (n > (uword)DiffPtrs(strend,strfree)) {

#ifdef FixedRegions
         if (qualfail)
            RetError(-304, nulldesc);
#endif                                  /* FixedRegions */

         RetError(-306, nulldesc);
         }
      }
   return Success;
   }
