/*
 * xwindow.h - macros and types used in Icon Windows.
 *
 * X provides a new definition of Success, so throw out Icon's Success.
 */
#undef Success
/*
 * DECstation X Window #includes check for #ifdef VMS, so undef it.
 */
#undef VMS
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
/*
 * Windows code prints lots of diagnostics if the symbol WINDEBUG is defined.
 * You must define DEBUG; level 0 means no debugging.
 */
#ifndef WINDEBUG
#define WINDEBUG 0
#endif

#ifndef MAXXOBJS
#define MAXXOBJS 256
#endif					/* MAXXOBJS */

/*
 * Macros to ease coding in which every X call must be done twice.
 */
#ifdef XBATCH
#define RENDER2(func,v1,v2) {\
   if(!ISBATCHMODE(w) && stdwin) \
      func(stddpy, stdwin, stdgc, v1, v2); \
   func(stddpy, stdpix, stdgc, v1, v2);}
#define RENDER3(func,v1,v2,v3) {\
   if(!ISBATCHMODE(w) && stdwin) \
      func(stddpy, stdwin, stdgc, v1, v2, v3); \
   func(stddpy, stdpix, stdgc, v1, v2, v3);}
#define RENDER4(func,v1,v2,v3,v4) {\
   if(!ISBATCHMODE(w) && stdwin) \
      func(stddpy, stdwin, stdgc, v1, v2, v3, v4); \
   func(stddpy, stdpix, stdgc, v1, v2, v3, v4);}
#define RENDER6(func,v1,v2,v3,v4,v5,v6) {\
   if(!ISBATCHMODE(w) && stdwin) \
      func(stddpy, stdwin, stdgc, v1, v2, v3, v4, v5, v6); \
   func(stddpy, stdpix, stdgc, v1, v2, v3, v4, v5, v6);}

#else				/* XBATCH */
#define RENDER2(func,v1,v2) {\
   if (stdwin) func(stddpy, stdwin, stdgc, v1, v2); \
   func(stddpy, stdpix, stdgc, v1, v2);}
#define RENDER3(func,v1,v2,v3) {\
   if (stdwin) func(stddpy, stdwin, stdgc, v1, v2, v3); \
   func(stddpy, stdpix, stdgc, v1, v2, v3);}
#define RENDER4(func,v1,v2,v3,v4) {\
   if (stdwin) func(stddpy, stdwin, stdgc, v1, v2, v3, v4); \
   func(stddpy, stdpix, stdgc, v1, v2, v3, v4);}
#define RENDER6(func,v1,v2,v3,v4,v5,v6) {\
   if (stdwin) func(stddpy, stdwin, stdgc, v1, v2, v3, v4, v5, v6); \
   func(stddpy, stdpix, stdgc, v1, v2, v3, v4, v5, v6);}
#endif				/* XBATCH */

/*
 * EVENT HANDLING
 *
 * Each window keeps an associated queue of events waiting to be
 * processed.  The queue consists mainly of characters, with
 * occasional mouse events thrown in.  The queue is an array of
 * unsigned shorts.
 *
 * Values 0..255 designate normal characters, and
 * are returned as strings of length 1.
 * Negative values designate special events and are returned as integers.
 * Values above 255 are pixel locations, and are translated down by 255
 * and then returned as integers.
 */

#define QUEUEUP(w,c) { \
  (w)->window->eventQueue[(w)->window->eQfront++] = c; \
  if ((w)->window->eQfront >= EQUEUELEN) (w)->window->eQfront = 0; \
}

#define QUEUEBACK(w,i) { (i)--; if ((i) < 0) (i) = EQUEUELEN-1; }


#define EQUEUELEN 256
#define MARGIN 0

/*
 * mode bits for the Icon window context (as opposed to X context)
 */

#define ISZOMBIE(w)     (WBITS(w) & 1)
#define ISCURSORON(w)   (WBITS(w) & 2)
#ifdef XBATCH
#define ISBATCHMODE(w)  (WBITS(w) & 4)
#endif				/* XBATCH */
#define ISREVERSE(w)    (CBITS(w) & 8)
#define ISUNDERLINE(w)  (CBITS(w) & 16)
#define ISBOXMODE(w)    (CBITS(w) & 32)
#define ISELLIPSE(w)    (CBITS(w) & 64)
#define ISDOUBLE(w)     (CBITS(w) & 128)
#define ISEXPOSED(w)    (WBITS(w) & 256)

#define SETZOMBIE(w)    (WBITS(w) |= 1)
#define SETCURSORON(w)  (WBITS(w) |= 2)
#ifdef XBATCH
#define SETBATCHMODE(w) (WBITS(w) |= 4)
#endif				/* XBATCH */
#define SETREVERSE(w)   (CBITS(w) |= 8)
#define SETUNDERLINE(w) (CBITS(w) |= 16)
#define SETBOXMODE(w)   (CBITS(w) |= 32)
#define SETELLIPSE(w)   (CBITS(w) |= 64)
#define SETDOUBLE(w)    (CBITS(w) |= 128)
#define SETEXPOSED(w)   (WBITS(w) |= 256)

#define CLRZOMBIE(w)    (WBITS(w) &= ~1)
#define CLRCURSORON(w)  (WBITS(w) &= ~2)
#ifdef XBATCH
#define CLRBATCHMODE(w) (WBITS(w) &= ~4)
#endif				/* XBATCH */
#define CLRREVERSE(w)   (CBITS(w) &= ~8)
#define CLRUNDERLINE(w) (CBITS(w) &= ~16)
#define CLRBOXMODE(w)   (CBITS(w) &= ~32)
#define CLRELLIPSE(w)   (CBITS(w) &= ~64)
#define CLRDOUBLE(w)    (CBITS(w) &= ~128)
#define CLREXPOSED(w)   (WBITS(w) &= ~256)


/*
 * Window Resources
 * Icon "Resources" are a layer on top of the window system resources,
 * provided in order to facilitate resource sharing and minimize the
 * number of calls to the window system.  Resources are reference counted.
 * These data structures are simple, static allocations of pointers
 * into dynamic, internal window system structures.
 *
 * The following constants define limitations in the system until this
 * code is rewritten to use dynamic allocation.
 */
#define DMAXCOLORS	256
#define WMAXCOLORS	256
#define MAXFONTS	24
#define MAXCOLORNAME	24
#define MAXFONTNAME	80
#define MAXDISPLAYNAME	24
#define MAXDISPLAYS	4
#define MAXBINDINGS	1024
#define MAXCONTEXTS	512
#define MAXWINDOWS	64
#define SHARED          0
#define MUTABLE         1
#define NUMCURSORSYMS	78


/*
 * Colors.  These are allocated within displays; they are currently
 * statically bounded to DMAXCOLORS colors per display.  Pointers
 * into the display's color table are also kept on a per-window
 * basis so that they may be (de)allocated when a window is cleared.
 * Colors are aliased by r,g,b value.  Allocations by name and r,g,b
 * share when appropriate.
 *
 * Color (de)allocation comprises a simple majority of the space
 * requirements of the current implementation.  A monochrome-only
 * version would take a lot less space.
 */
typedef struct {			/* ~40 bytes */
  int		refcount;
  char		name[MAXCOLORNAME];	/* name for XAttrib reads */
  unsigned short r, g, b;		/* rgb for colorsearch */
  unsigned long	c;			/* X pixel value */
  int           type;			/* SHARED or MUTABLE */
} wcolor, *wclrp;

/*
 * Fonts are allocated within displays.  MAXFONTS bounds the number of fonts
 * per display.  "Garbage" fonts are only freed if the font table fills up.
 */
typedef struct {			/* ~88 bytes */
  int		refcount;
  char		name[MAXFONTNAME];	/* name for XAttrib and fontsearch */
  int		height;			/* font height */
  XFontStruct *	fsp;			/* X font pointer */
} wfont, *wfp;

/*
 * Displays are maintained in a global array in rwinrsrc.c.
 */
typedef struct {			/* ~12K */
  int		refcount;
  char		name[MAXDISPLAYNAME];
  Display *	display;
  int		screen;
  int		numFonts;
  wfont		fonts[MAXFONTS];
  int           numColors;		/* allocated color info */
  wcolor	colors[DMAXCOLORS];
  Cursor	cursors[NUMCURSORSYMS];
} wdisplay, *wdp;
extern wdisplay wdsplys[];

/*
 * "Context" comprises the graphics context, and the font (i.e. text context).
 * Foreground and background colors (pointers into the display color table)
 * are stored here to reduce the number of window system queries.
 * Contexts are allocated out of a global array in rwinrsrc.c.
 */
typedef struct {			/* ~20 */
  int		refcount;
  GC		gc;			/* X graphics context */
  wclrp		fg, bg;
  wfp		font;
  int		fillstyle;
  int		linestyle;
  int		linewidth;
  int		drawop;
  int		bits;			/* context bits */
} wcontext, *wcp;

/*
 * "Window state" includes the actual X window and references to a large
 * number of resources allocated on a per-window basis.  Windows are
 * allocated out of a global array in rwinrsrc.c.  Windows remember the
 * first WMAXCOLORS colors they allocate, and deallocate them on clearscreen.
 */
typedef struct {			/* ~1.1K */
  int		refcount;		/* reference count */
  wdp		display;
  Window	win;			/* X window */
  int		ellipse_oldx;           /* used in ellipse outlines */
  int		y, x;			/* current cursor location, in pixels*/
  int		mousey,mousex;		/* current mouse location, in pixels */
  int		posy, posx;		/* desired upper lefthand corner */
  Pixmap	pix;			/* current screen state */
  Pixmap	prepix;			/* an initial image to display */
  int		height;			/* window height, in pixels */
  int		width;			/* window width, in pixels */
  int		pixheight;		/* pixmap height, in pixels */
  int		pixwidth;		/* pixmap width, in pixels */
  int		bits;			/* window bits */
  wclrp		winbg;			/* window background color */
  int           numColors;		/* allocated color info */
  short		theColors[WMAXCOLORS];	/* indices into display color table */
  int		theCursor;		/* index into cursor table */
  short		eventQueue[EQUEUELEN];  /* queue up received events */
  int		eQfront, eQback;

  long wmhintflags;			/* window manager hints */
  int initstate;
  int iconx, icony;
} wstate, *wsp;

/*
 * Icon window file variables are actually pointers to "bindings"
 * of a window and a context.  They are allocated out of a global
 * array in rwinrsrc.c.  There is one binding per Icon window value.
 */
typedef struct {			/* ~12 bytes */
  int refcount;
  wcp context;
  wsp window;
} wbinding, *wbp;
extern wbinding wbndngs[];

/*
 * Macros for resource access.  Since many attributes are buried
 * under several (possible window system specific) levels of
 * structures, macros are provided for convenient access.  The
 * root structure is always the window binding pointer wbp.
 */
#define FG(w)		((w)->context->fg->c)
#define FGNAME(w)	((w)->context->fg->name)
#define BG(w)		((w)->context->bg->c)
#define WBG(w)		((w)->window->winbg->c)
#define BGNAME(w)	((w)->context->bg->name)
#define FONT(w)		((w)->context->font->fsp)
#define FONTNAME(w)	((w)->context->font->name)
#define FHEIGHT(w)	((w)->context->font->height)
#define FASCENT(w)	((w)->context->font->fsp->max_bounds.ascent)
#define FDESCENT(w)	((w)->context->font->fsp->max_bounds.descent)
#define FWIDTH(w)	((w)->context->font->fsp->max_bounds.width)
#define DISPLAY(w)	((w)->window->display->display)
#define DISPLAYNAME(w)	((w)->window->display->name)
#define SCREEN(w)	((w)->window->display->screen)
#define COLORMAP(w)	DefaultColormap(DISPLAY(w),SCREEN(w))
#define XGC(w)		((w)->context->gc)
#define WIN(w)		((w)->window->win)
#define WIN_X(w)	((w)->window->x)
#define WIN_Y(w)	((w)->window->y)
#define MOUSE_X(w)	((w)->window->mousex)
#define MOUSE_Y(w)	((w)->window->mousey)
#define POS_X(w)	((w)->window->posx)
#define POS_Y(w)	((w)->window->posy)
#define PIX(w)		((w)->window->pix)
#define HEIGHT(w)	((w)->window->height)
#define WIDTH(w)	((w)->window->width)
#define PIXHEIGHT(w)	((w)->window->pixheight)
#define PIXWIDTH(w)	((w)->window->pixwidth)
#define CBITS(w)	((w)->context->bits)
#define WBITS(w)	((w)->window->bits)


/*
 * macros performing row/column to pixel y,x translations
 * computation is 1-based and depends on the current font's size.
 * exception: XTOCOL as defined is 0-based, because that's what its
 * clients seem to need.
 */
#define ROWTOY(w,row) ((row-1) * FHEIGHT(w) + FASCENT(w) + MARGIN)
#define COLTOX(w,col) ((col-1) * FWIDTH(w) + MARGIN)
#define YTOROW(w,y)   (((y) - MARGIN) / FHEIGHT(w))
#define XTOCOL(w,x)   (((x) - MARGIN) / FWIDTH(w))

wclrp alc_color			Params((wbp w,char *s));
wdp alc_display			Params((char *s));
wfp alc_font			Params((wbp w, char *s));
wcp alc_context			Params((wbp w));
wsp alc_winstate();
wbp alc_wbinding();

#define STDLOCALS(w) \
   wcp wc = (w)->context;\
   wsp ws = (w)->window; \
   GC      stdgc  = (wc)->gc; \
   Display *stddpy = (ws)->display->display; \
   Window  stdwin  = (ws)->win; \
   Pixmap  stdpix  = (ws)->pix

typedef struct {
  char *s;
  int i;
} stringint, *siptr;
extern stringint drawops[], cursorsyms[], attribs[];
extern char *si_i2s();
extern int si_s2i();

/*
 * Attributes
 */

#define A_BATCH		1
#define A_BG		2
#define A_BOX		3
#define A_COL		4
#define A_COLUMNS	5
#define A_CURSOR	6
#define A_DEPTH		7
#define A_DISPLAY	8
#define A_DISPLAYHEIGHT	9
#define A_DISPLAYWIDTH	10
#define A_DOUBLE	11
#define A_DRAWOP	12
#define A_ELLIPSE	13
#define A_FG		14
#define A_FHEIGHT	15
#define A_FILLSTYLE	16
#define A_FONT		17
#define A_FWIDTH	18
#define A_GEOMETRY	19
#define A_HEIGHT	20
#define A_ICONPOS	21
#define A_IMAGE		22
#define A_INITSTATE	23
#define A_LINES		24
#define A_LINESTYLE	25
#define A_LINEWIDTH	26
#define A_MOUSECOL	27
#define A_MOUSEROW	28
#define A_MOUSEX	29
#define A_MOUSEY	30
#define A_NORMAL	31
#define A_POINTER	32
#define A_POS		33
#define A_REVERSE	35
#define A_ROW		34
#define A_TEXTWIDTH	36
#define A_UNDERLINE	39
#define A_VISUAL	40
#define A_WIDTH		41
#define A_X		37
#define A_Y		38

#define NUMATTRIBS	41
