/*
 * mcontrol.c: main control of memory monitoring.
 */

#include "memmon.h"
#include "../h/monitor.h"

#undef EventMon			/* Use MemMon codes here, regardless */

hidden novalue gcmark Params((word n));

static units;
static pausereply = 0;
static initialized = 0;

/*
 *  skipgc(n) - skip the first n garbage collections.
 */
novalue skipgc(n)
int n;
   {
   int b, c;

   if (gclimit == 0)
      mquit(NormalExit);
   b = 0;
   while (n > 0)  {
      switch (c = getc(ifile)) {
         case EOF:			/* EOF */
            fprintf(stderr, "%s: hit EOF while skipping\n", progname);
            exit(ErrorExit);
            return;
         case '#':			/* comment */
         case ';':			/* pause */
            while ((c = getc(ifile)) != EOF && c != '\n')
               ;
            break;
         case '{':
            ncollect++;
            switch (b - '0') {
               case 4:  /* fall through -- old version of case 0 */
               case 0:  nexplicit++; break;
               case 1:  nstatic++;   break;
               case 2:  nstring++;   break;
               case 3:  nblock++;    break;
            }
            if (ncollect == gclimit)
               mquit(NormalExit);
            break;
         case '}':			/* end marking phase */
            n--;
            break;
         }
      b = c;
      }
   }

/*
 *  memmon() - main loop of the memory monitor.
 */
novalue memmon()
   {
   int cmd, prevcmd, c;
   word addr, len;
   int colr;
   char buf[LineSize], *msg;

   if (pauselimit == 0)
      return;

   prevcmd = 0;
   for (;;) {
      switch (cmd = getcmd(&addr, &len)) {
         case 0:			/* 0: end of file; terminate cleanly */
            if (!initialized) {
               fprintf(stderr, "%s: empty input file\n", progname);
               exit(ErrorExit);
               }
            if (prevcmd == '#')
               if (strncmp(buf, " Run-time error ", 16) == 0)
                  msg = strncpy (buf+12, "err", 3);
               else
                  msg = buf + 1;
            else if (prevcmd == '=')
               msg = "normal termination";
            else
               msg = "premature EOF";
            mpause('d', msg);
            mstatus(msg, C_Status);
            return;
         case '#':			/* #: comment */
            fgets(buf, LineSize, ifile);/* save text in buffer */
            if (!index(buf, '\n'))
               while ((c = getc(ifile)) != EOF && c != '\n')
                  ;
            break;
         case '<':			/* <: new memory layout */
            units = (len > 0) ? len : 4;/* set units if specified */
            getregion(&stc);
            getregion(&str);	
            getregion(&blk);	
            refresh();			/* redraw entire screen */
            str.used = 0;		/* will recalculate during marking */
            blk.used = 0;
            paintblk(&stc, (word)0, stc.length, C_Free);
            paintstr((word)0, str.length, C_Free, C_Bsep);
            paintblk(&blk, (word)0, blk.length, C_Free);
            if (ncollect > 0)
               mstatus("compacting", C_Status);
            initialized = 1;
            break;
         case '>':			/* >: new layout is complete */
            mstatus("running", C_Status);
            devflush();
            break;
         case '=':			/* =: check that we're in sync */
            rsync(&stc, "static");
            rsync(&str, "string");
            rsync(&blk, "blk");
            break;
   
         case E_String:			/* string allocation */
            paintstr(str.used, len, Unmarked + C_String, Unmarked + C_Ssep);
            str.used += len;
            break;
         case '$':			/* $: mmshow() of a string */
            colr = getshow();
            paintstr(addr, len, colr, Unmarked + C_Ssep);
            break;
   
         case E_Tvsubs:			/* Tvsubs; substring trapped var */
         case E_File:			/* T_File; file block */
         case E_Refresh:		/* T_Refresh; refresh block */
         case E_Lrgint:			/* T_Lrgint; long integer */
         case E_Real:			/* T_Real; real number */
         case E_Record:			/* T_Record; record block */
         case E_Set:			/* T_Set; set header block */
         case E_Selem:			/* Selem; set element block */
         case E_List:			/* T_List; list header block */
         case E_Lelem:			/* Lelem; list element block */
         case E_Table:			/* Table; table header block */
         case E_Telem:			/* Telem; table element block */
         case E_Slots:			/* T_Slots; hash buckets (slots) */
         case E_Tvtbl:			/* Tvtbl; table elem trapped var */
         case E_External:		/* T_External; external block */
         case E_Cset:			/* T_Cset; cset */
   
            len *= units;
            paintblk(&blk, blk.used, len, Unmarked + blkcolor[cmd]);
            blk.used += len;
            break;
   
         case '%':			/* %: mmshow() in the block region */
            addr *= units;
            len *= units;
            colr = getshow();
            paintblk(&blk, addr, len, colr);
            break;
   
         case E_Alien:			/* Alien block in static region */
         case E_Free:			/* Free block in static region */
            addr *= units;
            len *= units;
            paintblk(&stc, addr, len, blkcolor[cmd]);
            break;
   
         case E_Coexpr:			/* Co-expr block in static region */
            addr *= units;
            len *= units;
            paintblk(&stc, addr, len, Unmarked + blkcolor[cmd]);
            break;
   
         case 'Y':			/* Y: mmshow() in the static region */
            addr *= units;
            len *= units;
            colr = getshow();
            paintblk(&stc, addr, len, colr);
            break;
   
         case ';':			/* ;: mmpause() call */
            getc(ifile);		/* skip space character */
            fgets(buf, LineSize, ifile);/* read message */
            buf[strlen(buf)-1] = '\0';	/* remove newline */
            /* pause unless previous reply said "don't stop again" */
            if (pausereply != EOF && pausereply != 'g' && pausereply != 'G')
               pausereply = mpause('p', buf);
            break;
   
         case '{':			/* {: begin marking for garb. coll. */
            gcmark(len);
            break;
         case '!':			/* !: end garbage collection */
            gcwait('c', "end garbage collection");
            if (ncollect == gclimit) {
               mstatus("quit", C_Status);
               mquit(NormalExit);
               }
            mstatus("running", C_Status);
            break;
         default:
            fprintf(stderr, "%s: unexpected input char: %c\n", progname, cmd);
            exit(ErrorExit);
         }
      prevcmd = cmd;
      }
   }

/*
 * gcmark(n) - handle marking phase of garbage collection, reason n.
 */
static novalue gcmark(n)
word n;
   {
   word addr, len;
   int c, domark;
   char *s;

   if (index(whenpause, 'g') || index(whenpause, 'a') || index(miscflags, 'm'))
      domark = 1;
   else
      domark = 0;
   ncollect++;
   switch ((int)n) {
      case 4:  /* fall through -- old version of case 0 */
      case 0:  s = "collect(0) call";    nexplicit++; break;
      case 1:  s = "need static space";  nstatic++;   break;
      case 2:  s = "need string space";  nstring++;   break;
      case 3:  s = "need block space";   nblock++;    break;
      default: s = "g.c. reason lost";                break;
      }
   c = gcwait('f', s);
   if (c == '+' || c == '/')
      domark = 0;
   if (domark) {
      setmap(Unmarked, C_Unmarked);
      mstatus("marking", C_Status);
      }

   for (;;) switch (c = getcmd(&addr, &len)) {
      case '#':				/* #: comment */
         while ((c = getc(ifile)) != EOF && c != '\n')
            ;
         break;
      case 0:				/* 0: end of file (shouldn't happen) */
      case '}':				/* }: end marking phase */
         if (domark) {
            for (;;) {
               c = gcwait('a', "active data after marking");
               setmap(Marked, C_Marked);
               if (c == EOF || c == '+' || c == '/' || !index(whenpause, 'g'))
                  break;
               if (gcwait('g', "garbage to be discarded") != '-')
                  break;
               setmap(Unmarked, C_Unmarked);
               }
            }
         return;
      case E_String:				/* string allocation */
         if (domark)
            paintstr(addr, len, Marked + C_String, Marked + C_Ssep);
         break;

      case E_Tvsubs:			/* Tvsubs; substring trapped var */
      case E_File:			/* T_File; file block */
      case E_Refresh:			/* T_Refresh; refresh block */
      case E_Lrgint:			/* T_Lrgint; long integer */
      case E_Real:			/* T_Real; real number */
      case E_Record:			/* T_Record; record block */
      case E_Set:			/* T_Set; set header block */
      case E_Selem:			/* Selem; set element block */
      case E_List:			/* T_List; list header block */
      case E_Lelem:			/* Lelem; list element block */
      case E_Table:			/* T_Table; table header block */
      case E_Telem:			/* Telem; table element block */
      case E_Tvtbl:			/* Tvtbl; table elem trapped var */
      case E_Slots:			/* T_Slots; hash buckets (slots) */
      case E_External:			/* T_External; external block */
      case E_Cset:			/* T_Cset; cset */
         if (domark) {
            addr *= units;
            len *= units;
            paintblk(&blk, addr, len, Marked + blkcolor[c]);
            }
         break;

      case E_Alien:
      case E_Free:
      case E_Coexpr:
         if (domark) {
            addr *= units;
            len *= units;
            paintblk(&stc, addr, len, Marked + blkcolor[c]);
            }
         break;

      default:
         fprintf(stderr,"%s: unexpected input char during gc: %c\n",progname,c);
         exit(ErrorExit);
      }
   }

