/*
 * mmain.c: code shared among memmon back ends.
 */

/* in this file only, define globals instead of just declaring */
#define Global
#include "memmon.h"

hidden novalue mmoptions Params((int argc, char *argv[]));
hidden novalue mmusage Params((noargs));

static int justcolors = 0;		/* quit after setting colors? */
static int gcskip = 0;			/* number of gc's to skip */

/*
 * main program.
 */
novalue main(argc, argv)
int argc;
char *argv[];
   {
   progname = argv[0];			/* save program name for diagnostics */
   devsetup();				/* init device-dependent params */
   initcolors();			/* init color tables */
   mmoptions(argc, argv);		/* process command options */
   devinit();				/* open and initialize device */

   if (justcolors) {			/* if just color setting wanted: */
      setmap(Marked, C_Marked);			/* load color map */
      mquit(NormalExit);			/* terminate */
   }

   if (strcmp(whenpause,"") != 0 && strcmp(whenpause,"n") != 0 && !batchmode) {
      ttyi = fopen("/dev/tty", ReadText);	/* init tty input file */
      ttyo = fopen("/dev/tty", WriteText);	/* init tty output file */
      if (ttyi == NULL || ttyo == NULL) {
         fprintf(stderr, "%s: can't open /dev/tty; will not pause\n", progname);
         ttyi = ttyo = NULL;
         }
      }

   skipgc(gcskip);			/* skip the first n collections */
   memmon();				/* run the memory monitor */
   mquit(NormalExit);			/* terminate */
   }

/*
 * mmoptions(argc, argv) - process command options.
 */
static novalue mmoptions(argc, argv)
int argc;
char *argv[];
   {
   int c;
   extern char *optarg;	/* getopt() */
   extern int optindex;	/* getopt() */

   /*
    * set some defaults
    */
   gclimit = -1;			/* no limit on gc's */
   pauselimit = -1;			/* no limit on pauses */
   strcpy(whichregs, "sb");		/* regions */
   if (!whenpause[0]) {
      if (batchmode)
         strcpy(whenpause, "fagcpd");		/* pause/print points */
      else
         strcpy(whenpause, "fagcp");		/* pause/print points */
      }

   /*
    * We should ideally keep the next two lines in sync!
    * Unfortunately, there are too many options to include all in synopsis.
    */
#define MMUse "[-r {fsb}] [-p {fagcpdn}] [-bwhLMgqQS n] [-t title] [...] [file]"
   while ((c = getopt(argc, argv, "r:p:b:w:h:L:M:g:q:Q:S:t:mc:C:")) != EOF)
      switch (c) {
         case 'r':
            strncpy(whichregs, optarg, FlagSize);
            break;
         case 'p':
            strncpy(whenpause, optarg, FlagSize);
            break;
         case 'b':
            granularity = atoi(optarg);
            break;
         case 'w':
            width = atoi(optarg);
            break;
         case 'h':
            height = atoi(optarg);
            break;
         case 'L':
            textrow = atoi(optarg);
            break;
         case 'M':
            maxrow = atoi(optarg);
            break;
         case 'g':
            gcskip = atoi(optarg);
            break;
         case 'q':
            gclimit = atoi(optarg);
            break;
         case 'Q':
            pauselimit = atoi(optarg);
            break;
         case 'S':
            sfreq = atoi(optarg);
            break;
         case 't':
            title = optarg;
            break;
         case 'm':
            addflag(miscflags, 'm');
            break;
         case 'c':
            readcolors(optarg);
            break;
         case 'C':
            justcolors = 1;
            readcolors(optarg);
            break;
         default:
            mmusage();
         }

   if (optindex + 1 < argc)		/* if too many files named, quit */
      mmusage();

   if (optindex < argc) {			/* open input file, if specified */
      ifile = fopen(argv[optindex], ReadText);
      if (!ifile)
         pexit(argv[optindex]);
      if (!title)
         title = argv[optindex];		/* use file name as default title */
      }
   else					/* else use standard input */
      ifile = stdin;
   }

/*
 * mmusage() - diagnose bad memmon usage, and abort.
 */
static novalue mmusage()
   {
   fprintf(stderr, "usage: %s %s\n", progname, MMUse);
   exit(ErrorExit);
   }

/*
 * mquit(exitcode) - terminate the run.
 */
novalue mquit(exitcode)
int exitcode;
   {
   devterm();				/* close down the graphics device */
   exit(exitcode);
   }

/*
 * mpause(key, msg) - pause for response (or, in batchmode, print display).
 *
 * Skips the pause if a -p option didn't include the given key character.
 * Returns the last character before '\n', 0 if none, or EOF.
 */
int mpause(key, msg)
int key;
char *msg;
   {
   static int c1, c2;
   static int npauses = 0;


   if (!index(whenpause, key))		/* if no pause/print of this display */
      return 0;

   npauses++;				/* count this visit */

   if (batchmode) {			/* if batch, output snapshot */
      mstatus(msg, C_Prompt);		/* set prompt message */
      devsnap();			/* write frame */
      mstatus("", C_Background);	/* clear the prompt */
      if (npauses == pauselimit)	/* if limit hit, quit now */
         mquit(NormalExit);
      return 0;
      }

   if (ttyi == NULL || c1 == EOF)  {	/* if no tty, or EOF already seen */
      c1 = EOF;				/* fake EOF for return */
      return EOF;
      }
   else {
      fprintf(ttyo, "memmon pause: %s: ", msg);		/* issue prompt */
      fflush(ttyo);
      mstatus(msg, C_Prompt);				/* also post on screen*/
      devflush();
      c1 = 0;
      while ((c2 = getc(ttyi)) != '\n' && c2 != '\r')	/* get char up to CR */
         if (c2 == EOF || c2 == '\04')  {		/* handle eof */
            c1 = EOF;
            break;
            }
         else
            c1 = c2;					/* save previous char */
      }

   if (c1 == 'q' || npauses == pauselimit)
      mquit(NormalExit);		/* quit on 'q' or if limit hit */

   mstatus("", C_Background);		/* clear the prompt */
   return c1;				/* return char before CR */
   }
