/*
 * xmemmon.c: graphics driver for the X Window System  (R4 with Athena Widgets)
 *
 *  This file includes main() and therefore replaces mmain.c.
 */

#include <ctype.h>
#include <setjmp.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/MenuButton.h>

#define Global		/* instantiate globals */
#include "memmon.h"	/* cannot precede X includes (reason unknown) */

#define MMUse "[-r {fsb}] [-p {fagcpnr}] [-bwhLMg n] [-t title] [...] file"

#define PollFreq 250	/* poll for events after this many paints */
#define HSize 800	/* default horizontal screen size */
#define VSize 500	/* default vertical screen size */
#define RSize 20	/* default maximum row height */
#define TSize 20	/* default text row height */
#define Gran 4		/* default granularity */

#define ShowFlag 'm'	/* ShowMarking flag */
#define LoopFlag 'l'	/* Loop mode flag */

hidden novalue setinput	Params((noargs));
hidden novalue restart	Params((noargs));
hidden novalue initgc	Params((noargs));
hidden novalue makewids	Params((noargs));
hidden Widget  button	Params((char *label, int c, int active));
hidden novalue domenu	Params((Widget w,caddr_t clientdata,caddr_d calldata));
hidden novalue fixmenu	Params((noargs));
hidden novalue respond	Params((Widget w,caddr_t clientdata,caddr_d calldata));
hidden novalue act	Params((int c));
hidden novalue initfont	Params((int w));
hidden XFontStruct *pickfont Params((int w));
hidden novalue repaint	Params((int x, int y, int w, int h));
hidden novalue xpaint	Params((word s, word n, int c, int b));
hidden novalue retext	Params((noargs));
hidden novalue xtext	Params((char *s, int r, int c, int fg, int bg));
hidden novalue setpause	Params((int status));
hidden novalue xpoll	Params((noargs));

/* event action routines; must be void, not "novalue", for X macros to work */
hidden void click	Params((Widget w, XEvent *e, String *s, int n));
hidden void keypress	Params((Widget w, XKeyEvent *event));
hidden void configtxt	Params((Widget w, XConfigureEvent *event));
hidden void exposetxt	Params((Widget w, XExposeEvent *event));
hidden void configmem	Params((Widget w, XConfigureEvent *event));
hidden void exposemem	Params((Widget w, XExposeEvent *event));

/*
 * Window and Widget data.
 */
static Display *dpy;
static Screen *scrn;
static Window rootwin, txtwin, memwin;
static Widget topwidget, formwidget, buttons, text, memory;
static Widget menu;
static Widget restartbutton, stopbutton, stepbutton, runbutton, quitbutton;
static Pixmap checkmark;

/*
 * Color data.
 */
static Colormap xcmap;		/* X color map to use */
static int usecolor;		/* can we use color? */
static int chgcolor;		/* are the colors changeable? */
static XColor cell[MapSize];	/* specification for each color */
static GC gclist[MapSize];	/* GC for writing in each color */
static int background = -1;	/* overall background color */
static long black, white;	/* black and white pixel values */

/*
 * Window data buffers.
 */
static unsigned char *mbuf;	/* memory color buffer */
static word mbufsiz = 0;	/* current allocated size */

static XFontStruct *textfont;				/* font used for text */
static int textsize, textbase;				/* height, baseline */
static float charwidth;					/* width of one char */

static char tbuf[TextLines][TextLength + 1];		/* text chars */
static unsigned char tfg[TextLines][TextLength + 1];	/* foreground color */
static unsigned char tbg[TextLines][TextLength + 1];	/* background color */

/*
 * Miscellaneous globals.
 */

static int paused = TRUE;	/* are we paused? (first cleared by expose ev)*/
static int dostep = FALSE;	/* OK to single step once? */
static char lastact;		/* last action char (typein or equivalent) */

static jmp_buf jumpbuf;		/* env for longjumping on a restart */
static char titlebuf[TitleLength+1];	/* title buffer */

/*
 * Command options.
 */
static XrmOptionDescRec options[] = {
   /*
    * X-style options.
    */
   { "-width",		".memWidth",	XrmoptionSepArg, NULL },
   { "-height",		".memHeight",	XrmoptionSepArg, NULL },
   { "-granularity",	".granularity",	XrmoptionSepArg, NULL },
   { "-textHeight",	".textHeight",	XrmoptionSepArg, NULL },
   { "-rowHeight",	".rowHeight",	XrmoptionSepArg, NULL },
   { "-skipGC",		".skipGC",	XrmoptionSepArg, NULL },
   { "-title",		".title",	XrmoptionSepArg, NULL },
   { "-regions",	".regions",	XrmoptionSepArg, NULL },
   { "-pausePoints",	".pausePoints",	XrmoptionSepArg, NULL },
   { "-colorSpec",	".colorSpec",	XrmoptionSepArg, NULL },
   { "-showMarking",	".showMarking",	XrmoptionNoArg,  "True" },
   { "-loop",		".loop",	XrmoptionNoArg,  "True" },
   /*
    * Old-style options for backwards compatibility.
    *  (Note that the single character is not always the initial character.)
    */
   { "-w",		".memWidth",	XrmoptionSepArg, NULL },
   { "-h",		".memHeight",	XrmoptionSepArg, NULL },
   { "-b",		".granularity",	XrmoptionSepArg, NULL },
   { "-L",		".textHeight",	XrmoptionSepArg, NULL },
   { "-M",		".rowHeight",	XrmoptionSepArg, NULL },
   { "-g",		".skipGC",	XrmoptionSepArg, NULL },
   { "-t",		".title",	XrmoptionSepArg, NULL },
   { "-r",		".regions",	XrmoptionSepArg, NULL },
   { "-p",		".pausePoints",	XrmoptionSepArg, NULL },
   { "-c",		".colorSpec",	XrmoptionSepArg, NULL },
   { "-m",		".showMarking",	XrmoptionNoArg,  "True" },
   { "-l",		".loop",	XrmoptionNoArg,  "True" },
   };

/*
 * Application resource struct.
 *
 *  Variable names used here are externally visible (via Resource macro below)
 *  and so should follow X precedents and conventions when possible.
 */
typedef struct {
   int memWidth;
   int memHeight;
   int granularity;
   int textHeight;
   int rowHeight;
   int skipGC;
   String font;
   String title;
   String regions;
   String colorSpec;
   String pausePoints;
   Boolean showMarking;
   Boolean loop;
   } AppData, *AppDataPtr;

static AppData app_data;

/*
 * Application resource list.  Use no spaces in these macro calls.
 */
#define Resource(name,class,type,default) \
   {  Lit(name), Lit(class), Cat(XtR,type), sizeof(type), \
      XtOffset(AppDataPtr,name), XtRImmediate, (caddr_t)default  }
#define XtRint XtRInt

static XtResource resources[] = {
   Resource(memWidth,MemWidth,int,HSize),
   Resource(memHeight,MemHeight,int,VSize),
   Resource(granularity,Granularity,int,Gran),
   Resource(textHeight,TextHeight,int,TSize),
   Resource(rowHeight,RowHeight,int,RSize),
   Resource(skipGC,SkipGC,int,0),
   Resource(font,Font,String,NULL),
   Resource(title,Title,String,titlebuf),
   Resource(regions,Regions,String,"sb"),
   Resource(pausePoints,PausePoints,String,""),
   Resource(colorSpec,ColorSpec,String,NULL),
   Resource(showMarking,ShowMarking,Boolean,FALSE),
   Resource(loop,Loop,Boolean,FALSE),
   };

/*
 * Possible fonts, in order of preference.
 */

static char *fontlist[] = {
   "*-helvetica-bold-r-*--*-140-**",
   "*-helvetica-bold-r-*--*-120-**",
   "*-helvetica-medium-r-*--*-120-**",
   "*-helvetica-medium-r-*--*-100-**",
   "*-helvetica-medium-r-*--*-80-**",
   "*helvetica*",
   "*lucida*",
   "10x20",
   "9x15",
   "8x13",
   "7x13",
   "6x10",
   "5x8",
   NULL};

/*
 * Actions.
 */
static XtActionsRec actions[] = {
   { "ConfigTxt", configtxt },
   { "ExposeTxt", exposetxt },
   { "ConfigMem", configmem },
   { "ExposeMem", exposemem },
   { "Click",     click },
   { "KeyPress",  keypress },
   };

/*
 * Menu items.  Look at domenu() to see how each of these works.
 */
typedef struct {
   Widget wid;		/* widget corresponding to entry */
   char *flags;		/* address of flag string */
   char ch;		/* flag character */
   char *text;		/* menu entry label */
   } item;

static item itemlist[] = {
   { 0, whichregs, 'f', "show static region" },
   { 0, whichregs, 's', "show string region" },
   { 0, whichregs, 'b', "show block region" },
   { 0, 0, 0, 0, },
   { 0, whenpause, 'f', "pause when memory full" },
   { 0, whenpause, 'a', "pause showing active data" },
   { 0, whenpause, 'g', "pause showing garbage" },
   { 0, whenpause, 'c', "pause after compaction" },
   { 0, whenpause, 'p', "pause for mmpause() calls" },
   { 0, whenpause, 'r', "pause after a restart" },
   { 0, 0, 0, 0, },
   { 0, miscflags, ShowFlag, "always show marking" },
   { 0, miscflags, LoopFlag, "loop continuously" },
   };

/*
 * Data for the checkmark used with the menu.
 */
#define CkWidth 10
#define CkHeight 8
static char ckdata[] = {
   0x80, 0x03,
   0xC0, 0x01,
   0xE0, 0x00,
   0x72, 0x00,
   0x3B, 0x00,
   0x1E, 0x00,
   0x0C, 0x00,
   0x08, 0x00,
   };

/*
 * main program
 */
int main(argc, argv)
int argc;
char *argv[];
   {
   Dimension dwidth, dheight;

   progname = argv[0];		/* save program name */
   initcolors();		/* init default color table */
   gclimit = pauselimit = -1;	/* never quit without explicit action */

   /*
    * Initialize window system and process command options.
    */
   topwidget = XtInitialize("xmemmon", "XMemmon",
      options, XtNumber(options), &argc, argv);

   if (argc > 2 || (argc==2 && argv[1][0]=='-' && argv[1][1]!='\0')) {
      fprintf(stderr, "usage: %s %s\n", progname, MMUse);
      exit(ErrorExit);
      }
   setinput(argv[1]);		/* set up input file */

   /*
    * Move option values from app_data to their conventional places.
    */
   XtGetApplicationResources
      (topwidget, &app_data, resources, XtNumber(resources), NULL, 0);
   width = app_data.memWidth;
   height = app_data.memHeight;
   textrow = app_data.textHeight;
   maxrow = app_data.rowHeight;
   granularity = app_data.granularity;
   title = app_data.title;
   strncpy(whichregs, app_data.regions, FlagSize);
   strncpy(whenpause, app_data.pausePoints, FlagSize);
   if (app_data.showMarking)
      addflag(miscflags, ShowFlag);
   if (app_data.loop) {		/* loop mode: also set "-m" and "-p n" */
      addflag(miscflags, LoopFlag);
      addflag(miscflags, ShowFlag);
      strcpy(whenpause, "n");	
      }
   if (app_data.colorSpec)
      readcolors(app_data.colorSpec);

   /*
    * More window setup.
    */
   XtAddActions(actions, XtNumber(actions));
   dpy = XtDisplay(topwidget);
   scrn = XtScreen(topwidget);
   xcmap = DefaultColormapOfScreen(scrn);
   rootwin = RootWindowOfScreen(scrn);
   checkmark = XCreateBitmapFromData(dpy, rootwin, ckdata, CkWidth, CkHeight);

   initgc();			/* initialize graphics contexts */
   makewids();			/* create widgets */
   XtRealizeWidget(topwidget);	/* place them on the screen */
   xpoll();			/* wait for expose event */

   /*
    * Record the size of the window actually created.
    *  Select a text font.
    *  Configure pause points depending on the display type.
    */
   memwin = XtWindow(memory);
   txtwin = XtWindow(text);

   XtVaGetValues(memory,
      XtNwidth, &dwidth,
      XtNheight, &dheight,
      NULL);
   width = dwidth;
   height = dheight;

   initfont(width);		/* choose and initialize text font */

   if (!whenpause[0])
      strcpy(whenpause, "fagcpr");
   if (!chgcolor)		/* can't pause 'a' if can't change colormap */
      delflag(whenpause, 'a');

   /* 
    * And now, to work.
    */
   skipgc(app_data.skipGC);	/* skip initial GCs, if asked */
   setjmp(jumpbuf);		/* allow restart at this point */
   fixmenu();			/* make sure menu is right */
   memmon();			/* enter main loop */
   mquit(NormalExit);		/* quit by waiting for a click */
   /*NOTREACHED*/
   }

/*
 * restart() -- rewind the input file and start afresh.
 */
static novalue restart()
   {
   if (fseek(ifile, (long)0, 0) != 0) {
      mpause(0, "can't rewind");
      return;
      }
   /*
    * Force button redraw (else it gets missed after the longjmp).
    */
   XtVaSetValues (restartbutton, XtNsensitive, FALSE, NULL);
   XtVaSetValues (restartbutton, XtNsensitive, TRUE, NULL);
   /*
    * Restore the initial state and jump back to the top level.
    */
   XtVaSetValues (stepbutton, XtNsensitive, TRUE, NULL);
   clearerr(ifile);
   nstatic = nstring = nblock = nexplicit = 0;
   setmap(Marked, C_Marked);
   paused = FALSE;
   devpaint((word)0, mempixels, C_Free, -1);
   mpause('r', "ready");	/* pauses only if selected */
   longjmp(jumpbuf, 1);
   }

/*
 * setinput (fname) -- set input file; use stdin if null
 */
static novalue setinput(fname)
char *fname;
   {
   if (fname == NULL) {
      if (isatty(fileno(stdin))) {	/* catch common error */
         fprintf(stderr, "usage: %s %s\n", progname, MMUse);
         exit(ErrorExit);
         }
      ifile = stdin;
      strcpy(titlebuf, "stdin");
      }
   else if (strcmp(fname, "-") == 0) {
      /* allow a tty if explicit "-" given to mean stdin */
      ifile = stdin;
      strcpy(titlebuf, "stdin");
      }
   else {
      ifile = fopen(fname, ReadText);	/* open named file */
      if (!ifile)
         pexit(fname);
      strncpy(titlebuf, fname, TitleLength);
      }
   XtVaSetValues (topwidget,
      XtNtitle, titlebuf,
      XtNiconName, titlebuf,
      NULL);
   }

/*
 * mquit(exitcode) - terminate the run.
 */
novalue mquit(exitcode)
int exitcode;
   {
   if (exitcode != NormalExit)
      exit(exitcode);
   else
      for (;;)
         XtProcessEvent(XtIMAll);	/* wait forever for "quit" click */
   }

/*
 * initgc() - initialize graphics contexts.
 */
static novalue initgc()
   {
   int i, n, r, g, b;
   float lum;
   Visual *visual;
   XGCValues gcv;

   static char pattern[17][4] = {	/* 4x4 stipple patterns for monochrome*/
      /*  0 */	{  0,  0,  0,  0 },
      /*  1 */	{  0,  0,  2,  0 },
      /*  2 */	{  0,  1,  0,  4 },
      /*  3 */	{  0,  6,  2,  0 },
      /*  4 */	{ 10,  0,  5,  0 },
      /*  5 */	{  8,  4,  4,  3 },
      /*  6 */	{ 10,  4,  5,  2 },
      /*  7 */	{  8,  3,  5,  6 },
      /*  8 */	{ 10,  5, 10,  5 },
      /*  9 */	{  8, 15,  8,  7 },
      /* 10 */	{ 12,  7, 14,  3 },
      /* 11 */	{ 13,  7, 14,  3 },
      /* 12 */	{ 15,  5, 15, 10 },
      /* 13 */	{ 11, 13, 14, 15 },
      /* 14 */	{ 15, 11, 15, 14 },
      /* 15 */	{ 15, 15, 13, 15 },
      /* 16 */	{ 15, 15, 15, 15 },
   };

   black = BlackPixelOfScreen(scrn);
   white = WhitePixelOfScreen(scrn);

   if (DefaultDepthOfScreen(scrn) == 1) {
      usecolor = FALSE;		/* not color or grayscale */
      chgcolor = FALSE;
      }
   else {
      visual = DefaultVisualOfScreen(scrn);
      switch (visual->class) {
         case GrayScale:
         case PseudoColor:
         case DirectColor:
            usecolor = TRUE;
            chgcolor = TRUE;
            break;
         case StaticGray:
         case StaticColor:
         case TrueColor:
            usecolor = TRUE;
            chgcolor = FALSE;
            break;
         }
      }

   for (i = 0; i < MapSize; i++) {
      if (!chgcolor && i >= Marked)
         n = C_Marked;
      else
         n = i % NColors;

      r = cmap[n].red;
      g = cmap[n].green;
      b = cmap[n].blue;
      cell[i].red   = 0x101 * r;
      cell[i].green = 0x101 * g;
      cell[i].blue  = 0x101 * b;
      cell[i].flags = DoRed | DoGreen | DoBlue;
      if (!usecolor) {
         /*
          * On a b/w screen, approximate the graylevel using a stipple pattern.
          */
         lum = (.30 * r + .59 * g + .11 * b) / 255.0;
         g = (int)((XtNumber(pattern) - .001) * lum);
         gcv.foreground = white;
         gcv.background = black;
         gcv.fill_style = FillOpaqueStippled;
         gcv.stipple = XCreatePixmapFromBitmapData
            (dpy, rootwin, pattern[g], 4, 4, 1, 0, 1);
         gclist[i] = XCreateGC(dpy, rootwin,
            GCForeground | GCBackground | GCFillStyle | GCStipple,
            &gcv);
         }
      else if (!chgcolor || (i < NColors)) {
         /*
          * This is a fixed color that will not need to change.
          *  (There are other fixed colors.  If we could easily identify
          *  them we would need fewer private colors and could run more
          *  copies of xmemmon simultaneously.)
          */
         if (!XAllocColor(dpy, xcmap, &cell[i])) {
            fprintf(stderr, "%s: can't allocate color cells.\n", progname);
            exit(ErrorExit);
            }
         gcv.foreground = cell[i].pixel;
         gclist[i] = XCreateGC(dpy, rootwin, GCForeground, &gcv);
         }
      else {
         /*
          * This color will change, so we need a writable entry.
          */
         if (!XAllocColorCells(dpy, xcmap, FALSE, NULL, 0, &cell[i].pixel, 1)) {
            fprintf(stderr,
               "%s: can't allocate enough private colors.\n", progname);
            exit(ErrorExit);
            }
         gcv.foreground = cell[i].pixel;
         gclist[i] = XCreateGC(dpy, rootwin, GCForeground, &gcv);
         }
      }
   }

/*
 * makewids() -- create widgets
 */
static novalue makewids()
   {
   int i;
   item *m;

   /*
    * The form widget contains and orgainzes the other widgets
    */
   formwidget = XtVaCreateManagedWidget("form", formWidgetClass, topwidget,
      XtNcolormap, xcmap,		/* for inheritance by descendents */
      XtNbackground, cell[C_Background].pixel,
      NULL);

   /*
    * Text area.
    */
   text = XtVaCreateManagedWidget("legend", coreWidgetClass, formwidget,
      XtNtop, XtChainTop,
      XtNbottom, XtChainTop,
      XtNleft, XtChainLeft,
      XtNright, XtChainRight,
      XtNwidth, width,
      XtNheight, TextLines * textrow,
      XtNborderWidth, (usecolor ? 0 : 1),
      XtNbackground, cell[C_Background].pixel,
      XtNtranslations, XtParseTranslationTable("\
         <Expose>:	ExposeTxt()\n\
         <Configure>:	ConfigTxt()\n\
         <Btn1Up>:	Click(\"\n\")\n\
         <Btn2Up>:	Click(\" \")\n\
         <Btn3Down>:	XawPositionSimpleMenu(menu) MenuPopup(menu)\n\
         <KeyPress>:	KeyPress()\n\
         "),
      NULL);

   /*
    * Memory area.
    */
   memory = XtVaCreateManagedWidget("memory", coreWidgetClass, formwidget,
      XtNfromVert, text,
      XtNtop, XtChainTop,
      XtNbottom, XtChainBottom,
      XtNleft, XtChainLeft,
      XtNright, XtChainRight,
      XtNwidth, width,
      XtNheight, height,
      XtNborderWidth, (usecolor ? 0 :	1),
      XtNbackground, cell[C_Background].pixel,
      XtNtranslations, XtParseTranslationTable("\
         <Expose>:	ExposeMem()\n\
         <Configure>:	ConfigMem()\n\
         <Btn1Up>:	Click(\"\n\")\n\
         <Btn2Up>:	Click(\" \")\n\
         <Btn3Down>:	XawPositionSimpleMenu(menu) MenuPopup(menu)\n\
         <KeyPress>:	KeyPress()\n\
         "),
      NULL);

   /*
    * Button row.
    */
   buttons = XtVaCreateManagedWidget("buttons", formWidgetClass, formwidget,
      XtNfromVert, memory,
      XtNleft, XtChainLeft,
      XtNright, XtChainLeft,
      XtNtop, XtChainBottom,
      XtNbottom, XtChainBottom,
      XtNborderWidth, 0,
      XtNdefaultDistance, 0, 
      XtNbackground, cell[C_Background].pixel,
      NULL);
   restartbutton = button("restart", 'r', TRUE);
   stopbutton = button("stop", '.', FALSE);
   stepbutton = button("step", ' ', TRUE);
   runbutton = button("run", '\n', FALSE);
   quitbutton = button("quit", 'q', TRUE);

   /*
    * Menu.
    */
   menu = XtCreatePopupShell("menu", simpleMenuWidgetClass, topwidget, NULL, 0);
   for (i = 0; i < XtNumber(itemlist); i++) {
      m = &itemlist[i];
      if (m->ch == 0)
         m->wid = XtCreateManagedWidget("---",smeLineObjectClass,menu,NULL,0);
      else if (m->ch != 'a' || chgcolor) {	/* can't do 'a' if !chgcolor */
         m->wid = XtVaCreateManagedWidget(m->text, smeBSBObjectClass, menu,
            XtNleftMargin, CkWidth + 10,
            XtNrightMargin, CkWidth + 10,
            NULL);
         XtAddCallback(m->wid, XtNcallback, domenu, (caddr_t)m);
         }
      }

   }

/*
 * button (label, ch, active) -- create button widget.
 *
 *  Each button widget calls "respond" with the char "ch" as client_data.
 */
static Widget button(label, ch, active)
char *label;
int ch, active;
   {
   static Widget w = NULL;	/* remembers previous (adjacent) button */
   static int distance = 0;	/* distance from adjacent button */

   w = XtVaCreateManagedWidget(label, commandWidgetClass, buttons,
      XtNfromHoriz, w,
      XtNhorizDistance, distance,
      XtNleft, XtChainRight,
      XtNright, XtChainRight,
      XtNtop, XtChainTop,
      XtNbottom, XtChainTop,
      XtNsensitive, active,
      XtNborderColor, (usecolor ? cell[C_Status].pixel : black),
      XtNforeground, (usecolor ? cell[C_Status].pixel : black),
      XtNbackground, (usecolor ? cell[C_Background].pixel : white),
      NULL);
   XtAddCallback(w, XtNcallback, respond, (caddr_t)ch);
   distance = 4;		/* all but first are spaced over a bit */
   return w;
   }

/*
 * domenu(w, client_data, call_data) - respond to menu selection.
 */
/*ARGSUSED*/
static novalue domenu(w, client_data, call_data)
Widget w;
caddr_t client_data, call_data;
   {
   item *m = (item *)client_data;

   /*
    * Add or remove the flag character for this button from its flag string.
    */
   if (index(m->flags, m->ch)) {
      delflag(m->flags, m->ch);			/* flag was set; clear it */
      XtVaSetValues(w, XtNleftBitmap, None, NULL);
      }
   else {
      addflag(m->flags, m->ch);			/* flag was clear; set it */
      XtVaSetValues(w, XtNleftBitmap, checkmark, NULL);
      }
   /*
    * If all three regions are deselected, memmon will bomb on the next
    *  refresh.  Prevent this by reselecting a different region.
    */
   if (m->flags == whichregs && whichregs[0] == '\0') {
      if (m->ch == 'b')
         strcpy(whichregs, "s");	/* deselected block, select string */
      else
         strcpy(whichregs, "b");	/* else select block region */
      fixmenu();
      }
   /*
    * Checking "loop continuously" also sets and clears other flags.
    */
   if (m->flags == miscflags && m->ch == LoopFlag) {
      if (index(miscflags, LoopFlag)) {
         /*
          * We have just set the loop flag.  Clear pauses, restart if at EOF.
          */
         addflag(miscflags, ShowFlag);
         strcpy(whenpause, "n");
         setpause(FALSE);
         if (feof(ifile))
            restart();
         }
      else {
         /*
          * We just cleared the loop flag; pause but don't change anything else
          */
         setpause(TRUE);
         }
      fixmenu();
      }
   }

/*
 * fixmenu() - make sure the check marks on the menu are correct.
 */
static novalue fixmenu()
   {
   item *m;

   for (m = itemlist; m < itemlist + XtNumber(itemlist); m++)
      if (m->wid != NULL && m->flags != NULL)
         XtVaSetValues(m->wid, XtNleftBitmap,
            (index(m->flags, m->ch) ? checkmark : None), NULL);
   }

/*
 * Simple event handlers.
 */

/*ARGSUSED*/
static void click(w, e, s, n)		/* mouse button click */
Widget w;
XButtonEvent *e;
String *s;
int n;
   {
   if (n > 0 && **s != '\0')
      act(**s);
   }

/*ARGSUSED*/
static void keypress(w, event)		/* keypress event */
Widget w;
XKeyEvent *event;
   {
   char buffer[4];

   if (XLookupString(event, buffer, sizeof(buffer), NULL, NULL) == 1)
      act(buffer[0]);
   }

/*ARGSUSED*/
static novalue respond(w, client_data, call_data)	/* button widget press*/
Widget w;
caddr_t client_data, call_data;
   {
   char c = (char)client_data;

   act(c);
   }

/*
 * act(c) - react to keypress c or the equivalent button or mouse event
 */
static novalue act(c)
int c;
   {
   lastact = c;
   switch (c) {

      /*
       * RUN, Left Button, Return, Linefeed, '/', '+', '-'
       */
      case '\n':
      case '\r':
      case '/':
      case '+':
      case '-':
         setpause(FALSE);
         break;

      /*
       * STEP, Middle Button, Space Bar
       */
      case ' ':
         if (paused)
            dostep = TRUE;
         else
            setpause(TRUE);
         break;

      /*
       * STOP
       */
      case '.':
         setpause(TRUE);
         break;

      /*
       * 'g'
       */
      case 'g':
         strcpy(whenpause, "n");	/* clear all pause flags */
         fixmenu();
         lastact = '\n';		/* fake \n typein */
         setpause(FALSE);		/* clear current pause */
         break;

      /*
       * QUIT, 'q', DEL, ^C
       */
      case '\177':	/* DEL */
      case '\03':	/* ^C */
      case 'q':
         exit(NormalExit);
         break;

      /*
       * RESTART, 'r'
       */
      case 'r':
         restart();
         break;
      }
   }

/*
 * initfont(w) - initialize text font for window of width w.
 */
static novalue initfont (w)
int w;
   {
   int i;
   XGCValues gcv;
   
   textfont = pickfont(w);	/* choose font */
   textsize = textfont->ascent + textfont->descent;
   textbase = (textrow - textsize) / 2 + textfont->ascent;

   for (i = 0; i < MapSize; i++) {
      gcv.font = textfont->fid;
      XChangeGC(dpy, gclist[i], GCFont, &gcv);
      }
   }

/*
 * pickfont(w) - choose a font based on width w;
 */
static XFontStruct *pickfont(w)
int w;
   {
   XFontStruct *font = NULL;
   char *teststring = "coexpr";		/* a good string for testing font size*/
   int slen;				/* its length */
   int maxwidth;			/* max width available for its display*/
   char **p, *fontname, *bestname; 
   int n, bestwidth;

   /*
    * If the user specified a font, use it regardless of its size;
    */
   if (app_data.font) {
      if ((font = XLoadQueryFont(dpy, app_data.font)) != NULL)
         return font;
      fprintf(stderr, "%s: can't load font %s\n", progname, app_data.font);
      app_data.font = NULL;
      }

   /*
    * Find the first font from the list that isn't too big to fit.
    */
   slen = strlen(teststring);
   maxwidth = slen * w / TextLength;
   bestname = NULL;
   bestwidth = 10000; /* huge */
   p = fontlist;
   while ((fontname = *p++) != NULL) {
      if ((font = XLoadQueryFont(dpy, fontname)) != NULL) {
         n = XTextWidth(font, teststring, slen);
         if (n < maxwidth)		/* it fits -- use it! */
            return font;
         if (n < bestwidth) {		/* remember best seen */
            bestname = fontname;
            bestwidth = n;
            }
         XFreeFont(dpy, font);
         font = NULL;
         }
      }

   /*
    * Nothing was tiny enough. Pick the closest match.
    */
   if (bestname != NULL)
      font = XLoadQueryFont (dpy, bestname);
   if (font != NULL)
      return font;
   fprintf(stderr, "%s: can't find any font\n", progname);
   exit(ErrorExit);
   /*NOTREACHED*/
   }

/*
 * devmap() - load color map into device.
 */
novalue devmap()
   {
   int i;

   if (!chgcolor)		/* if we can't change color map, return */
      return;

   for (i = 0; i < MapSize; i++) {
      cell[i].red   = 0x101 * cmap[i].red;
      cell[i].green = 0x101 * cmap[i].green;
      cell[i].blue  = 0x101 * cmap[i].blue;
      cell[i].flags = DoRed | DoGreen | DoBlue;
      }
   /*
    *  Don't try to alter the first NColors; they are fixed.
    */
   XStoreColors(dpy, xcmap, cell + NColors, MapSize - NColors);
   }

/*
 * devflood(c) - fill screen with color c.
 */
novalue devflood(c)
int c;
   {
   word n;

   XFillRectangle(dpy, memwin, gclist[c], 0, 0, width, height);
   XFillRectangle(dpy, txtwin, gclist[c], 0, 0, width, 2 * textrow);

   n = (word)height * (word)width;	/* max display w/o new refresh */
   if (n > mbufsiz) {
      mbuf = (unsigned char *)XtRealloc(mbuf, n);
      mbufsiz = n;
      }

   memfill((char *)mbuf, c, (word)mbufsiz);
   memfill((char *)tbuf, 0, (word)(TextLines * TextLength));
   memfill((char *)tbg, c, (word)(TextLines * TextLength));
   background = c;
   charwidth = (float)width / (float)TextLength;
   }

/*
 * devsnap() - batch mode snapshot; no action needed here.
 */
novalue devsnap()
   {
   }

/*
 * devflush() - flush output.
 */
novalue devflush()
   {
   XFlush(dpy);
   }

/*
 * devpaint(start, n, color, b) - paint n pixels in given color.
 *  If b >= 0, the last pixel is to be that color instead (for a border)
 */
novalue devpaint(s, n, c, b)
word s, n;
int c, b;
   {
   unsigned char *p;
   static int unpolled;

   /*
    * paint the screen
    */
   if (b >= 0)				/* if border, decr total count */
      n--;
   xpaint(s, n, c, b);			/* paint the real window */
   
   /*
    * remember colors in memory buffer
    */
   p = mbuf + s;
   while (n--)				/* fill pixels */
      *p++ = c;
   if (b >= 0)				/* if border, set its value */
      *p++ = b;

   /*
    * poll for events if needed, and occasionally if not
    */

   if (paused || (++unpolled >= PollFreq)) {
      xpoll();
      unpolled = 0;
      }
   }

/*
 * repaint a specified rectangle of memory on the screen.
 */
static novalue repaint (x, y, w, h)
int x, y, w, h;
   {
   unsigned char *p, *q;
   word row, lastrow;
   word s, n;
   int b, c;

   XFillRectangle(dpy, memwin, gclist[background], x, y, w, h);
   row = y / memrow;
   lastrow = (y + h - 1) / memrow;
   while (row <= lastrow) {
      s = row++ * width + x;
      if (s > mbufsiz)
         break;
      p = mbuf + s;
      q = p + w;
      while (p < q) {
         s = p - mbuf;
         c = *p++;
         n = 1;
         while (p < q && *p == c)
            n++, p++;
         if (p < q && p[0] != p[1])
            b = *p++;
         else
            b = -1;
         xpaint(s, n, c, b);
         }
      }
   }

/*
 * xpaint (start, n, color, b) - paint n pixels, plus one more if b!=0
 */
static novalue xpaint(s, n, c, b)
word s, n;
int c, b;
   {
   word x, y, h;

   x = s % width;			/* where on row? */
   y = memrow * (s / width);		/* which row? */
   h = memrow - 1;			/* height of row */
   if (h < 1)
      h = 1;

   while (x + n > width)  {		/* draw all rows but last */
      XFillRectangle(dpy, memwin, gclist[c], x, y, width - x, h);
      n -= width - x;				/* decr count */
      x = 0;					/* move to start of next row */
      y += memrow;
      }
   if (n) {				/* last row */
      XFillRectangle(dpy, memwin, gclist[c], x, y, n, h);
      x += n;
      }
   if (b >= 0)  {			/* border */
      XDrawLine(dpy, memwin, gclist[b], x, y, x, y + h - 1); 
      }
   }

/*
 * exposemem - redraw parts of memory in response to an expose event
 */
/*ARGSUSED*/
static void exposemem(w, event)
Widget w;
XExposeEvent *event;
   {
   if (mbufsiz == 0) {	/* if zero, must be first exposure event */
      setpause(FALSE);	/* clear pause flag to start display */
      return;		/* and with mbufsiz == 0, there's nothing to draw */
      }
   repaint(event->x, event->y, event->width, event->height);
   }

/*
 * configmem - layout and redraw memory in response to a resize event
 */
/*ARGSUSED*/
static void configmem(w, event)
Widget w;
XConfigureEvent *event;
   {
   word nbytes, newpixels;
   int newrows, oldw, oldh;

   if (width == event->width && height == event->height)
      return;
   oldw = width;
   oldh = height;
   width = event->width;
   height = event->height;
   memheight = height;
    
   nbytes = stc.displ + str.displ + blk.displ;

   /* calc ideal scaling */
   newpixels = (nbytes + granularity - 1) / granularity;
   newrows = (newpixels + width - 1) / width;

   /* set scaling, but no more than command option maximum */
   memrow = height / newrows;
   if (memrow > maxrow)
      memrow = maxrow;

   if (memrow <= 0)			/* sanity check, and extreme case */
      memrow = 1;
   if (newrows > height)
      newrows = height;

   /* set total number of pixels and minimum y value */
   mempixels = newrows * width;
   if (mempixels > newpixels)
      mempixels = newpixels;
   ymin = height - newrows * memrow;
   if (ymin < 0)
      ymin = 0;

   /*
    * Only repaint the intersection of the old and new windows;
    *  a separate expose event has been generated and will repaint the rest.
    */
   repaint(0,0, ((oldw<width) ? oldw : width), ((oldh<height) ? oldh : height));
   }

/*
 * devtext(string, row, col, fgcolr, bgcolr) - save and output text.
 */
novalue devtext(s, row, col, fg, bg)
char *s;
int row, col, fg, bg;
   {
   int n;

   n = strlen(s);
   if (col + n > TextLength)
      n = TextLength - col;

   memfill((char *)&tfg[row][col], fg, (word)(n + 1));	/* save for exposure */
   memfill((char *)&tbg[row][col], bg, (word)(n + 1));
   strncpy(&tbuf[row][col], s, n + 1);

   xtext (s, row, col, fg, bg);
   }

/*
 * retext() - redraw the entire text section (without clearing background).
 */
static novalue retext()
   {
   int c, row, col, k;

   /*
    * identify and display individual text strings
    */
   for (row = 0; row < TextLines; row++)  {	/* for each text line */
      col = 0;
      while (col < TextLength)  {
         if (tbuf[row][col] == '\0' || tfg[row][col] == tbg[row][col])
            col++;			/* skip unused text positions */
         else {
            for (k = col;
               k < TextLength && tfg[row][k]==tfg[row][col] && tbuf[row][k];
               k++)
                  ;			/* find run of one color */
            c = tbuf[row][k];
            tbuf[row][k] = '\0';	/* temporarily terminate string */
            xtext(&tbuf[row][col], row, col,
               (int)tfg[row][col], (int)tbg[row][col]);
            tbuf[row][k] = c;	/* restore text buffer */
            col = k;
            }
         }
      }
   }

/*
 * xtext(string, row, col, fgcolr, bgcolr) - output text.
 *
 *  Assumes charwidth has been set to the desired size.
 */
static novalue xtext(s, row, col, fg, bg)
char *s;
int row, col, fg, bg;
   {
   int x, y, n;

   x = col * charwidth;				/* where on line? */
   y = textrow * row;				/* which line? */
   n = strlen(s);
   XFillRectangle(dpy, txtwin, gclist[bg], x, y,
      (int)(charwidth * (n + 1)) - 2, textrow - 2); /* draw background */

   if (n == 0)					/* if no text atop background */
      return;

   /* we draw string even if fg=bg in case it extends beyond box limits */
   XDrawString(dpy, txtwin, gclist[fg], x+(int)(charwidth/2), y+textbase, s, n);
   }

/*
 * exposetxt() - redraw text section in response to expose event.
 */
/*ARGSUSED*/
static void exposetxt(w, event)
Widget w;
XExposeEvent *event;
   {
   if (background < 0)				
      return;				/* nothing to do yet */
   /*
    * Redraw the portion of the background that needs refreshing.
    */
   XFillRectangle(dpy, txtwin, gclist[background],
      event->x, event->y, event->width, event->height);
   /*
    * It's easiest to just redraw all of the text strings.
    */
   retext();
   }

/*
 * configtxt - recalc char width and redraw legend in response to a resize.
 */
/*ARGSUSED*/
static void configtxt(w, event)
Widget w;
XConfigureEvent *event;
   {
   if (app_data.font == NULL) {	/* pick new font if not user specified */
      XFreeFont(dpy, textfont);
      initfont(event->width);
      }
   XFillRectangle(dpy, txtwin, gclist[background],
      event->x, event->y, event->width, event->height);
   charwidth = (float)event->width / (float)TextLength;
   retext();
   }

/*
 * mpause(ch, msg) - pause for response.
 *
 *  Skip the pause, returning 0, if the pauseflags don't include ch; else
 *  return the character (or char equivalent of a button) that acks the pause.
 */
int mpause(ch, msg)
int ch;
char *msg;
   {
   if (ch != 0) {	/* if a conditional pause (not an error notification) */

      if (feof(ifile)) {
         /*
          * This is the EOF pause.  If loopmode, force a restart.
          *  Otherwise just turn off "STOP", "STEP, and "RUN".
          */
         if (index(miscflags, LoopFlag))	/* if loop flag set */
            restart();
         XtVaSetValues (stopbutton, XtNsensitive, FALSE, NULL);
         XtVaSetValues (stepbutton, XtNsensitive, FALSE, NULL);
         XtVaSetValues (runbutton, XtNsensitive, FALSE, NULL);
         return 0;
      }
   
      if (!index(whenpause, ch)) {
         /*
          * We don't want to pause for this event; so just return.
          */
         setpause(FALSE);
         return 0;
         }
      }
   
   mstatus(msg, C_Prompt);		/* post the reason */
   setpause(TRUE);
   xpoll();				/* wait for acknowledgement */
   mstatus("", C_Background);		/* clear the prompt */
   return lastact;			/* return action character */
   }

/*
 * setpause(s) - set pause status on or off, but only if not at EOF.
 */
static novalue setpause(s)
int s;
   {
   if (feof(ifile))
      return;
   paused = s;
   XtVaSetValues (stopbutton, XtNsensitive, !s, NULL);
   XtVaSetValues (runbutton, XtNsensitive, s, NULL);
   }

/*
 * xpoll() - check for and handle any pending events.  
 *
 *  If paused, waits until pause is cleared or a single-step is given.
 */
static novalue xpoll()
   {
   while (XtPending() || (paused && !dostep))
      XtProcessEvent(XtIMAll);
   dostep = FALSE;
   }
